import logging
from pathlib import Path
from typing import Any, Dict, List, Optional

from knot_resolver.constants import WATCHDOG_LIB
from knot_resolver.datamodel import KresConfig
from knot_resolver.manager.config_store import ConfigStore, only_on_real_changes_update
from knot_resolver.manager.triggers import cancel_cmd, trigger_cmd, trigger_renew

logger = logging.getLogger(__name__)

FilesToWatch = Dict[Path, Optional[str]]


def watched_files_config(config: KresConfig) -> List[Any]:
    return [
        config.network.tls.watchdog,
        config.network.tls.cert_file,
        config.network.tls.key_file,
        config.local_data.rpz,
    ]


if WATCHDOG_LIB:
    from watchdog.events import (
        FileSystemEvent,
        FileSystemEventHandler,
    )
    from watchdog.observers import Observer

    class FilesWatchdogEventHandler(FileSystemEventHandler):
        def __init__(self, files: FilesToWatch, config: KresConfig) -> None:
            self._files = files
            self._config = config

        def _trigger(self, cmd: Optional[str]) -> None:
            if cmd:
                trigger_cmd(self._config, cmd)
            trigger_renew(self._config)

        def on_created(self, event: FileSystemEvent) -> None:
            src_path = Path(str(event.src_path))
            if src_path in self._files.keys():
                logger.info(f"Watched file '{src_path}' has been created")
                self._trigger(self._files[src_path])

        def on_deleted(self, event: FileSystemEvent) -> None:
            src_path = Path(str(event.src_path))
            if src_path in self._files.keys():
                logger.warning(f"Watched file '{src_path}' has been deleted")
                cmd = self._files[src_path]
                if cmd:
                    cancel_cmd(cmd)
            for file in self._files.keys():
                if file.parent == src_path:
                    logger.warning(f"Watched directory '{src_path}' has been deleted")
                    cmd = self._files[file]
                    if cmd:
                        cancel_cmd(cmd)

        def on_moved(self, event: FileSystemEvent) -> None:
            src_path = Path(str(event.src_path))
            if src_path in self._files.keys():
                logger.info(f"Watched file '{src_path}' has been moved")
                self._trigger(self._files[src_path])

        def on_modified(self, event: FileSystemEvent) -> None:
            src_path = Path(str(event.src_path))
            if src_path in self._files.keys():
                logger.info(f"Watched file '{src_path}' has been modified")
                self._trigger(self._files[src_path])

    _files_watchdog: Optional["FilesWatchdog"] = None

    class FilesWatchdog:
        def __init__(self, files_to_watch: FilesToWatch, config: KresConfig) -> None:
            self._observer = Observer()

            event_handler = FilesWatchdogEventHandler(files_to_watch, config)
            dirs_to_watch: List[Path] = []
            for file in files_to_watch.keys():
                if file.parent not in dirs_to_watch:
                    dirs_to_watch.append(file.parent)

            for d in dirs_to_watch:
                self._observer.schedule(
                    event_handler,
                    str(d),
                    recursive=False,
                )
                logger.info(f"Directory '{d}' scheduled for watching")

        def start(self) -> None:
            self._observer.start()

        def stop(self) -> None:
            self._observer.stop()
            self._observer.join()


@only_on_real_changes_update(watched_files_config)
async def _init_files_watchdog(config: KresConfig, force: bool = False) -> None:
    if WATCHDOG_LIB:
        global _files_watchdog

        if _files_watchdog:
            _files_watchdog.stop()
        files_to_watch: FilesToWatch = {}

        # network.tls
        if config.network.tls.watchdog and config.network.tls.cert_file and config.network.tls.key_file:
            net_tls = f"net.tls('{config.network.tls.cert_file}', '{config.network.tls.key_file}')"
            files_to_watch[config.network.tls.cert_file.to_path()] = net_tls
            files_to_watch[config.network.tls.key_file.to_path()] = net_tls

        # local-data.rpz
        if config.local_data.rpz:
            for rpz in config.local_data.rpz:
                if rpz.watchdog:
                    files_to_watch[rpz.file.to_path()] = None

        if files_to_watch:
            logger.info("Initializing files watchdog")
            _files_watchdog = FilesWatchdog(files_to_watch, config)
            _files_watchdog.start()


async def init_files_watchdog(config_store: ConfigStore) -> None:
    # register files watchdog callback
    await config_store.register_on_change_callback(_init_files_watchdog)



import logging
from typing import Dict, Optional

from knot_resolver.controller.interface import KresID
from knot_resolver.controller.registered_workers import command_registered_workers, get_registered_workers_kresids
from knot_resolver.datamodel import KresConfig
from knot_resolver.utils.modeling.parsing import DataFormat

logger = logging.getLogger(__name__)


async def collect_kresd_workers_metrics(config: KresConfig) -> Optional[Dict[KresID, object]]:
    if config.monitoring.metrics == "manager-only":
        logger.debug("Skipping kresd stat collection due to configuration")
        return None

    cmd = "collect_statistics()"
    if config.monitoring.metrics == "lazy":
        cmd = "collect_lazy_statistics()"
    logger.debug(f"Collecting stats from all kresd workers using method '{cmd}'")

    return await command_registered_workers(cmd)


async def report_json(config: KresConfig) -> bytes:
    metrics_raw = await collect_kresd_workers_metrics(config)
    metrics_dict: Dict[str, Optional[object]] = {}

    if metrics_raw:
        for kresd_id, kresd_metrics in metrics_raw.items():
            metrics_dict[str(kresd_id)] = kresd_metrics
    else:
        # if we have no metrics, return None for every kresd worker
        for kresd_id in get_registered_workers_kresids():
            metrics_dict[str(kresd_id)] = None

    return DataFormat.JSON.dict_dump(metrics_dict).encode()



# noqa: INP001
import argparse
import json
import os
import sys
from pathlib import Path
from typing import List, Optional, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, comp_get_words, register_command
from knot_resolver.utils import which
from knot_resolver.utils.requests import request

PROCS_TYPE = List


@register_command
class DebugCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        self.proc_type: Optional[str] = namespace.proc_type
        self.sudo: bool = namespace.sudo
        self.gdb: str = namespace.gdb
        self.print_only: bool = namespace.print_only
        self.gdb_args: List[str] = namespace.extra if namespace.extra is not None else []
        super().__init__(namespace)

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        debug = subparser.add_parser(
            "debug",
            help="Run GDB on the manager's subprocesses",
        )
        debug.add_argument(
            "--sudo",
            dest="sudo",
            help="Run GDB with sudo",
            action="store_true",
            default=False,
        )
        debug.add_argument(
            "--gdb",
            help="Custom GDB executable (may be a command on PATH, or an absolute path)",
            type=str,
            default=None,
        )
        debug.add_argument(
            "--print-only",
            help="Prints the GDB command line into stderr as a Python array, does not execute GDB",
            action="store_true",
            default=False,
        )
        debug.add_argument(
            "proc_type",
            help="Optional, the type of process to debug. May be 'kresd', 'gc', or 'all'.",
            choices=["kresd", "gc", "all"],
            type=str,
            nargs="?",
            default="kresd",
        )
        return debug, DebugCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:  # noqa: C901, PLR0912, PLR0915
        if self.gdb is None:
            try:
                gdb_cmd = str(which.which("gdb"))
            except RuntimeError:
                print("Could not find 'gdb' in $PATH. Is GDB installed?", file=sys.stderr)
                sys.exit(1)
        elif "/" not in self.gdb:
            try:
                gdb_cmd = str(which.which(self.gdb))
            except RuntimeError:
                print(f"Could not find '{self.gdb}' in $PATH.", file=sys.stderr)
                sys.exit(1)
        else:
            gdb_cmd_path = Path(self.gdb).absolute()
            if not gdb_cmd_path.exists():
                print(f"Could not find '{self.gdb}'.", file=sys.stderr)
                sys.exit(1)
            gdb_cmd = str(gdb_cmd_path)

        response = request(args.socket, "GET", f"processes/{self.proc_type}")
        if response.status != 200:
            print(response, file=sys.stderr)
            sys.exit(1)

        procs = json.loads(response.body)
        if not isinstance(procs, PROCS_TYPE):
            print(
                f"Unexpected response type '{type(procs).__name__}' from manager. Expected '{PROCS_TYPE.__name__}'",
                file=sys.stderr,
            )
            sys.exit(1)
        if len(procs) == 0:
            print(
                f"There are no processes of type '{self.proc_type}' available to debug",
                file=sys.stderr,
            )

        exec_args = []

        # Put `sudo --` at the beginning of the command.
        if self.sudo:
            try:
                sudo_cmd = str(which.which("sudo"))
            except RuntimeError:
                print("Could not find 'sudo' in $PATH. Is sudo installed?", file=sys.stderr)
                sys.exit(1)
            exec_args.extend([sudo_cmd, "--"])

        # Attach GDB to processes - the processes are attached using the `add-inferior` and `attach` GDB
        # commands. This way, we can debug multiple processes.
        exec_args.extend([gdb_cmd, "--"])
        exec_args.extend(["-init-eval-command", "set detach-on-fork off"])
        exec_args.extend(["-init-eval-command", "set schedule-multiple on"])
        exec_args.extend(["-init-eval-command", f'attach {procs[0]["pid"]}'])
        inferior = 2
        for proc in procs[1:]:
            exec_args.extend(["-init-eval-command", "add-inferior"])
            exec_args.extend(["-init-eval-command", f"inferior {inferior}"])
            exec_args.extend(["-init-eval-command", f'attach {proc["pid"]}'])
            inferior += 1

        num_inferiors = inferior - 1
        if num_inferiors > 1:
            # Now we switch back to the first process and add additional provided GDB arguments.
            exec_args.extend(["-init-eval-command", "inferior 1"])
            exec_args.extend(
                [
                    "-init-eval-command",
                    "echo \\n\\nYou are now debugging multiple Knot Resolver processes. To switch between "
                    "them, use the 'inferior <n>' command, where <n> is an integer from 1 to "
                    f"{num_inferiors}.\\n\\n",
                ]
            )
        exec_args.extend(self.gdb_args)

        if self.print_only:
            print(f"{exec_args}")
        else:
            os.execl(*exec_args)



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from knot_resolver.controller.interface import KresID
    from knot_resolver.datamodel.config_schema import KresConfig
    from knot_resolver.manager.config_store import ConfigStore


LOGGING_LEVEL_STARTUP = logging.DEBUG
PID_FILE_NAME = "knot-resolver.pid"

FIX_COUNTER_ATTEMPTS_MAX = 2
FIX_COUNTER_DECREASE_INTERVAL_SEC = 30 * 60
PROCESSES_WATCHDOG_INTERVAL_SEC: float = 5


def kres_cache_dir(config: "KresConfig") -> Path:
    return config.cache.storage.to_path()


def policy_loader_config_file(_config: "KresConfig") -> Path:
    return Path("policy-loader.conf")


def kresd_config_file(_config: "KresConfig", kres_id: "KresID") -> Path:
    return Path(f"kresd{int(kres_id)}.conf")


def kresd_config_file_supervisord_pattern(_config: "KresConfig") -> Path:
    return Path("kresd%(process_num)d.conf")


def supervisord_config_file(_config: "KresConfig") -> Path:
    return Path("supervisord.conf")


def supervisord_config_file_tmp(_config: "KresConfig") -> Path:
    return Path("supervisord.conf.tmp")


def supervisord_pid_file(_config: "KresConfig") -> Path:
    return Path("supervisord.pid")


def supervisord_sock_file(_config: "KresConfig") -> Path:
    return Path("supervisord.sock")


def supervisord_subprocess_log_dir(_config: "KresConfig") -> Path:
    return Path("logs")


class _UserConstants:
    """Class for accessing constants, which are technically not constants as they are user configurable."""

    def __init__(self, config_store: "ConfigStore", working_directory_on_startup: str) -> None:
        self._config_store = config_store
        self.working_directory_on_startup = working_directory_on_startup


_user_constants: Optional[_UserConstants] = None


async def init_user_constants(config_store: "ConfigStore", working_directory_on_startup: str) -> None:
    global _user_constants
    _user_constants = _UserConstants(config_store, working_directory_on_startup)


def user_constants() -> _UserConstants:
    assert _user_constants is not None
    return _user_constants



import base64
import json
from hashlib import blake2b
from typing import Any


def structural_etag(obj: Any) -> str:
    m = blake2b(digest_size=15)
    m.update(json.dumps(obj, sort_keys=True).encode("utf8"))
    return base64.urlsafe_b64encode(m.digest()).decode("utf8")



import errno
import socket
import sys
from http.client import HTTPConnection
from typing import Any, Literal, Optional
from urllib.error import HTTPError, URLError
from urllib.parse import quote, unquote, urlparse
from urllib.request import AbstractHTTPHandler, Request, build_opener, install_opener, urlopen


class SocketDesc:
    def __init__(self, socket_def: str, source: str) -> None:
        self.source = source
        if ":" in socket_def:
            # `socket_def` contains a schema, probably already URI-formatted, use directly
            self.uri = socket_def
        else:
            # `socket_def` is probably a path, convert to URI
            self.uri = f'http+unix://{quote(socket_def, safe="")}'

        while self.uri.endswith("/"):
            self.uri = self.uri[:-1]


class Response:
    def __init__(self, status: int, body: str) -> None:
        self.status = status
        self.body = body

    def __repr__(self) -> str:
        return f"status: {self.status}\nbody:\n{self.body}"


def _print_conn_error(error_desc: str, url: str, socket_source: str) -> None:
    host: str
    try:
        parsed_url = urlparse(url)
        host = unquote(parsed_url.hostname or "(Unknown)")
    except Exception as e:
        host = f"(Invalid URL: {e})"
    msg = f"""
{error_desc}
\tURL:           {url}
\tHost/Path:     {host}
\tSourced from:  {socket_source}
Is the URL correct?
\tUnix socket would start with http+unix:// and URL encoded path.
\tInet sockets would start with http:// and domain or ip
    """
    print(msg, file=sys.stderr)


def request(
    socket_desc: SocketDesc,
    method: Literal["GET", "POST", "HEAD", "PUT", "DELETE"],
    path: str,
    body: Optional[str] = None,
    content_type: str = "application/json",
) -> Response:
    while path.startswith("/"):
        path = path[1:]
    url = f"{socket_desc.uri}/{path}"
    req = Request(
        url,
        method=method,
        data=body.encode("utf8") if body is not None else None,
        headers={"Content-Type": content_type},
    )
    # req.add_header("Authorization", _authorization_header)

    timeout_m = 5  # minutes
    try:
        with urlopen(req, timeout=timeout_m * 60) as response:
            return Response(response.status, response.read().decode("utf8"))
    except HTTPError as err:
        return Response(err.code, err.read().decode("utf8"))
    except URLError as err:
        if err.errno == errno.ECONNREFUSED or isinstance(err.reason, ConnectionRefusedError):
            _print_conn_error("Connection refused.", url, socket_desc.source)
        elif err.errno == errno.ENOENT or isinstance(err.reason, FileNotFoundError):
            _print_conn_error("No such file or directory.", url, socket_desc.source)
        else:
            _print_conn_error(str(err), url, socket_desc.source)
        sys.exit(1)
    except (TimeoutError, socket.timeout):
        _print_conn_error(
            f"Connection timed out after {timeout_m} minutes."
            "\nIt does not mean that the operation necessarily failed."
            "\nSee Knot Resolver's log for more information.",
            url,
            socket_desc.source,
        )
        sys.exit(1)


# Code heavily inspired by requests-unixsocket
# https://github.com/msabramo/requests-unixsocket/blob/master/requests_unixsocket/adapters.py
class UnixHTTPConnection(HTTPConnection):
    def __init__(self, unix_socket_url: str, timeout: float = 60) -> None:
        """
        Create an HTTP connection to a unix domain socket.

        :param unix_socket_url: A URL with a scheme of 'http+unix' and the
        netloc is a percent-encoded path to a unix domain socket. E.g.:
        'http+unix://%2Ftmp%2Fprofilesvc.sock/status/pid'
        """
        super().__init__("localhost", timeout=timeout)
        self.unix_socket_path = unix_socket_url
        self.timeout = timeout
        self.sock: Optional[socket.socket] = None

    def __del__(self) -> None:  # base class does not have d'tor
        if self.sock:
            self.sock.close()

    def connect(self) -> None:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect(self.unix_socket_path)
        self.sock = sock


class UnixHTTPHandler(AbstractHTTPHandler):
    def __init__(self) -> None:
        super().__init__()

        def open_(self: UnixHTTPHandler, req: Any) -> Any:
            return self.do_open(UnixHTTPConnection, req)  # type: ignore[arg-type]

        setattr(UnixHTTPHandler, "http+unix_open", open_)
        setattr(UnixHTTPHandler, "http+unix_request", AbstractHTTPHandler.do_request_)


opener = build_opener(UnixHTTPHandler())
install_opener(opener)



from typing import List, Optional

from knot_resolver.datamodel.types import IPv6Network, IPv6Network96, TimeUnit
from knot_resolver.utils.modeling import ConfigSchema


class Dns64Schema(ConfigSchema):
    """
    DNS64 (RFC 6147) configuration.

    ---
    enable: Enable/disable DNS64.
    prefix: IPv6 prefix to be used for synthesizing AAAA records.
    reverse_ttl: TTL in CNAME generated in the reverse 'ip6.arpa.' subtree.
    exclude_subnets: IPv6 subnets that are disallowed in answer.
    """

    enable: bool = False
    prefix: IPv6Network96 = IPv6Network96("64:ff9b::/96")
    reverse_ttl: Optional[TimeUnit] = None
    exclude_subnets: Optional[List[IPv6Network]] = None



import functools
import os
from pathlib import Path


@functools.lru_cache(maxsize=16)
def which(binary_name: str) -> Path:
    """
    Search $PATH and return the absolute path of that executable.

    The results of this function are LRU cached.

    If not found, throws an RuntimeError.
    """
    possible_directories = os.get_exec_path()
    for dr in possible_directories:
        p = Path(dr, binary_name)
        if p.exists():
            return p.absolute()

    raise RuntimeError(f"Executable {binary_name} was not found in $PATH")



# ruff: noqa: E501
import logging
import os
import socket
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from knot_resolver.constants import API_SOCK_FILE, RUN_DIR, VERSION, WORKERS_SUPPORT
from knot_resolver.datamodel.cache_schema import CacheSchema
from knot_resolver.datamodel.defer_schema import DeferSchema
from knot_resolver.datamodel.dns64_schema import Dns64Schema
from knot_resolver.datamodel.dnssec_schema import DnssecSchema
from knot_resolver.datamodel.forward_schema import FallbackSchema, ForwardSchema
from knot_resolver.datamodel.globals import Context, get_global_validation_context, set_global_validation_context
from knot_resolver.datamodel.local_data_schema import LocalDataSchema, RPZSchema, RuleSchema
from knot_resolver.datamodel.logging_schema import LoggingSchema
from knot_resolver.datamodel.lua_schema import LuaSchema
from knot_resolver.datamodel.management_schema import ManagementSchema
from knot_resolver.datamodel.monitoring_schema import MonitoringSchema
from knot_resolver.datamodel.network_schema import NetworkSchema
from knot_resolver.datamodel.options_schema import OptionsSchema
from knot_resolver.datamodel.rate_limiting_schema import RateLimitingSchema
from knot_resolver.datamodel.templates import KRESD_CONFIG_TEMPLATE, POLICY_LOADER_CONFIG_TEMPLATE
from knot_resolver.datamodel.types import EscapedStr, IntPositive, WritableDir
from knot_resolver.datamodel.view_schema import ViewSchema
from knot_resolver.utils.modeling import ConfigSchema
from knot_resolver.utils.modeling.base_schema import lazy_default
from knot_resolver.utils.modeling.exceptions import AggregateDataValidationError, DataValidationError

WORKERS_MAX = 256

logger = logging.getLogger(__name__)


def _cpu_count() -> Optional[int]:
    try:
        return len(os.sched_getaffinity(0))  # type: ignore[attr-defined]
    except (NotImplementedError, AttributeError) as e:
        cpus = os.cpu_count()
        if cpus is None:
            logger.warning(
                "The number of usable CPUs could not be determined using"
                f" 'os.sched_getaffinity()' or 'os.cpu_count()':\n{e}"
            )
        return cpus


def workers_max_count() -> int:
    c = _cpu_count()
    if c:
        return c * 10
    return WORKERS_MAX


def _get_views_tags(views: List[ViewSchema]) -> List[str]:
    tags = []
    for view in views:
        if view.tags:
            tags += [str(tag) for tag in view.tags if tag not in tags]
    return tags


def _check_local_data_tags(
    views_tags: List[str], rules_or_rpz: Union[List[RuleSchema], List[RPZSchema]]
) -> Tuple[List[str], List[DataValidationError]]:
    tags = []
    errs = []

    i = 0
    for rule in rules_or_rpz:
        tags_not_in = []
        if rule.tags:
            for tag in rule.tags:
                tag_str = str(tag)
                if tag_str not in tags:
                    tags.append(tag_str)
                if tag_str not in views_tags:
                    tags_not_in.append(tag_str)
            if len(tags_not_in) > 0:
                errs.append(
                    DataValidationError(
                        f"some tags {tags_not_in} not found in '/views' tags", f"/local-data/rules[{i}]/tags"
                    )
                )
            i += 1
    return tags, errs


class KresConfig(ConfigSchema):
    class Raw(ConfigSchema):
        """
        Knot Resolver declarative configuration.

        ---
        version: Version of the configuration schema. By default it is the latest supported by the resolver, but couple of versions back are be supported as well.
        nsid: Name Server Identifier (RFC 5001) which allows DNS clients to request resolver to send back its NSID along with the reply to a DNS request.
        hostname: Internal DNS resolver hostname. Default is machine hostname.
        rundir: Directory where the resolver can create files and which will be it's cwd.
        workers: The number of running kresd (Knot Resolver daemon) workers. If set to 'auto', it is equal to number of CPUs available.
        management: Configuration of management HTTP API.
        options: Fine-tuning global parameters of DNS resolver operation.
        network: Network connections and protocols configuration.
        views: List of views and its configuration.
        local_data: Local data for forward records (A/AAAA) and reverse records (PTR).
        forward: List of Forward Zones and its configuration.
        fallback: Config for fallback on resolution failure.
        cache: DNS resolver cache configuration.
        dnssec: DNSSEC configuration.
        dns64: DNS64 (RFC 6147) configuration.
        logging: Logging and debugging configuration.
        monitoring: Metrics exposition configuration (Prometheus, Graphite)
        lua: Custom Lua configuration.
        rate_limiting: Configuration of rate limiting.
        defer: Configuration of request prioritization (defer).
        """

        version: int = 1
        nsid: Optional[EscapedStr] = None
        hostname: Optional[EscapedStr] = None
        rundir: WritableDir = lazy_default(WritableDir, str(RUN_DIR))
        workers: Union[Literal["auto"], IntPositive] = IntPositive(1)
        management: ManagementSchema = lazy_default(ManagementSchema, {"unix-socket": str(API_SOCK_FILE)})
        options: OptionsSchema = OptionsSchema()
        network: NetworkSchema = NetworkSchema()
        views: Optional[List[ViewSchema]] = None
        local_data: LocalDataSchema = LocalDataSchema()
        forward: Optional[List[ForwardSchema]] = None
        fallback: FallbackSchema = FallbackSchema()
        cache: CacheSchema = lazy_default(CacheSchema, {})
        dnssec: DnssecSchema = DnssecSchema()
        dns64: Dns64Schema = Dns64Schema()
        logging: LoggingSchema = LoggingSchema()
        monitoring: MonitoringSchema = MonitoringSchema()
        rate_limiting: RateLimitingSchema = RateLimitingSchema()
        defer: DeferSchema = DeferSchema()
        lua: LuaSchema = LuaSchema()

    _LAYER = Raw

    #### When ADDING options, please also update config_nodes() in ../manager/manager.py
    nsid: Optional[EscapedStr]
    hostname: EscapedStr
    rundir: WritableDir
    workers: IntPositive
    management: ManagementSchema
    options: OptionsSchema
    network: NetworkSchema
    views: Optional[List[ViewSchema]]
    local_data: LocalDataSchema
    forward: Optional[List[ForwardSchema]]
    fallback: FallbackSchema
    cache: CacheSchema
    dnssec: DnssecSchema
    dns64: Dns64Schema
    logging: LoggingSchema
    monitoring: MonitoringSchema
    rate_limiting: RateLimitingSchema
    defer: DeferSchema
    lua: LuaSchema

    def _hostname(self, obj: Raw) -> Any:
        if obj.hostname is None:
            return socket.gethostname()
        return obj.hostname

    def _workers(self, obj: Raw) -> Any:
        no_workers_support_msg = (
            "On this system, you cannot run more than one worker because "
            "SO_REUSEPORT (Linux) or SO_REUSEPORT_LB (FreeBSD) socket option is not supported."
        )
        if not WORKERS_SUPPORT and (int(obj.workers) > 1):
            raise ValueError(no_workers_support_msg)

        if obj.workers == "auto":
            if not WORKERS_SUPPORT:
                logger.info(
                    "Running on system without support for multiple workers,"
                    f"' workers' configuration automatically set to 1. {no_workers_support_msg}"
                )
                return IntPositive(1)

            count = _cpu_count()
            if count:
                return IntPositive(count)
            raise ValueError(
                "The number of available CPUs to automatically set the number of running 'kresd' workers could not be determined."
                "The number of workers can be configured manually in 'workers' option."
            )

        return obj.workers

    def _validate(self) -> None:  # noqa: C901
        # warn about '/management/unix-socket' not located in '/rundir'
        if self.management.unix_socket and self.management.unix_socket.to_path().parent != self.rundir.to_path():
            logger.warning(
                f"The management API unix-socket '{self.management.unix_socket}'"
                f" is not located in the resolver's rundir '{self.rundir}'."
                " This can lead to permissions issues."
            )

        # enforce max-workers config
        workers_max = workers_max_count()
        if int(self.workers) > workers_max:
            raise ValueError(
                f"can't run with more workers than the recommended maximum {workers_max} or hardcoded {WORKERS_MAX}"
            )

        # sanity check
        cpu_count = _cpu_count()
        if cpu_count and int(self.workers) > 10 * cpu_count:
            raise ValueError(
                "refusing to run with more then 10 workers per cpu core, the system wouldn't behave nicely"
            )

        # get all tags from views
        views_tags = []
        if self.views:
            views_tags = _get_views_tags(self.views)

        # get local-data tags and check its existence in views
        errs = []
        local_data_tags = []
        if self.local_data.rules:
            rules_tags, rules_errs = _check_local_data_tags(views_tags, self.local_data.rules)
            errs += rules_errs
            local_data_tags += rules_tags
        if self.local_data.rpz:
            rpz_tags, rpz_errs = _check_local_data_tags(views_tags, self.local_data.rpz)
            errs += rpz_errs
            local_data_tags += rpz_tags

        # look for unused tags in /views
        unused_tags = views_tags.copy()
        for tag in local_data_tags:
            if tag in unused_tags:
                unused_tags.remove(tag)
        if len(unused_tags) > 1:
            errs.append(DataValidationError(f"unused tags {unused_tags} found", "/views"))

        # raise all validation errors
        if len(errs) == 1:
            raise errs[0]
        if len(errs) > 1:
            raise AggregateDataValidationError("/", errs)

    def render_kresd_lua(self) -> str:
        # FIXME the `cwd` argument is used only for configuring control socket path
        # it should be removed and relative path used instead as soon as issue
        # https://gitlab.nic.cz/knot/knot-resolver/-/issues/720 is fixed
        return KRESD_CONFIG_TEMPLATE.render(cfg=self, cwd=os.getcwd())

    def render_policy_loader_lua(self) -> str:
        return POLICY_LOADER_CONFIG_TEMPLATE.render(cfg=self, cwd=os.getcwd())


def get_rundir_without_validation(data: Dict[str, Any]) -> WritableDir:
    """
    Without fully parsing, try to get a rundir from a raw config data, otherwise use default.

    Attempts a dir validation to produce a good error message.
    Used for initial manager startup.
    """
    return WritableDir(data["rundir"] if "rundir" in data else str(RUN_DIR), object_path="/rundir")


def kres_config_json_schema() -> Dict[str, Any]:
    """
    At this moment, to create any instance of 'ConfigSchema' even with default values, it is necessary to set the global context.

    In the case of generating a JSON schema, strict validation must be turned off, otherwise it may happen that the creation of the JSON schema fails,
    It may fail due to non-existence of the directory/file or their rights.
    This should be fixed in the future. For more info, see 'datamodel.globals.py' module.
    """
    context = get_global_validation_context()
    set_global_validation_context(Context(None, False))

    schema = KresConfig.json_schema(
        schema_id=f"https://www.knot-resolver.cz/documentation/v{VERSION}/_static/config.schema.json",
        title="Knot Resolver configuration JSON schema",
        description=f"Version Knot Resolver {VERSION}",
    )
    # setting back to previous values
    set_global_validation_context(context)

    return schema



from typing import List

from knot_resolver import KresBaseError


class KresSubprocessControllerError(KresBaseError):
    """Class for errors that are raised in the controller module."""


class KresSubprocessControllerExec(Exception):  # noqa: N818
    """
    Custom non-error exception that indicates the need for exec().

    Raised by the controller (supervisord) and caught by the controlled process (manager).
    The exception says that the process needs to perform a re-exec during startup.
    This ensures that the process runs under the controller (supervisord) in a process tree hierarchy.
    """

    def __init__(self, exec_args: List[str], *args: object) -> None:
        self.exec_args = exec_args
        super().__init__(*args)



from typing import Literal, Union

from knot_resolver.datamodel.types import DomainName, EscapedStr, IPAddress, PortNumber, TimeUnit
from knot_resolver.utils.modeling import ConfigSchema


class GraphiteSchema(ConfigSchema):
    enable: bool = False
    host: Union[None, IPAddress, DomainName] = None
    port: PortNumber = PortNumber(2003)
    prefix: EscapedStr = EscapedStr("")
    interval: TimeUnit = TimeUnit("5s")
    tcp: bool = False

    def _validate(self) -> None:
        if self.enable and not self.host:
            raise ValueError("'host' option must be configured to enable graphite bridge")


class MonitoringSchema(ConfigSchema):
    """
    ---
    metrics: configures, whether metrics/statistics will be collected by the resolver
    graphite: optionally configures where should graphite metrics be sent to
    """  # noqa: D205, D400, D415

    metrics: Literal["manager-only", "lazy", "always"] = "lazy"
    graphite: GraphiteSchema = GraphiteSchema()



# noqa: INP001
import argparse
import sys
from typing import List, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, register_command
from knot_resolver.utils.requests import request


@register_command
class ReloadCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)
        self.force: bool = namespace.force

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        reload = subparser.add_parser(
            "reload",
            help="Tells the resolver to reload YAML configuration file."
            " Old processes are replaced by new ones (with updated configuration) using rolling restarts."
            " So there will be no DNS service unavailability during reload operation.",
        )
        reload.add_argument(
            "--force",
            help="Force a reload, even if the configuration hasn't changed.",
            action="store_true",
            default=False,
        )
        return reload, ReloadCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return {}

    def run(self, args: CommandArgs) -> None:
        response = request(args.socket, "POST", "reload/force" if self.force else "reload")

        if response.status != 200:
            print(response, file=sys.stderr)
            sys.exit(1)



"""
Effectively the same as normal __main__.py.

However, we moved it's content over to this
file to allow us to exclude the __main__.py file from black's autoformatting
"""

import argparse
import sys
from typing import NoReturn

from knot_resolver.constants import CONFIG_FILE, VERSION
from knot_resolver.manager.logger import logger_startup
from knot_resolver.manager.server import start_server
from knot_resolver.utils import compat


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Knot Resolver - caching DNS resolver")
    parser.add_argument(
        "-V",
        "--version",
        help="Get version",
        action="version",
        version=VERSION,
    )
    parser.add_argument(
        "-c",
        "--config",
        help="One or more configuration files to load."
        f" Overrides default configuration file location at '{str(CONFIG_FILE)}'"
        " Files must not contain the same options."
        " However, they may extend individual subsections."
        " The location of the first configuration file determines"
        "the prefix for every relative path in the configuration.",
        type=str,
        nargs="+",
        required=False,
        default=[str(CONFIG_FILE)],
    )
    return parser.parse_args()


def main() -> NoReturn:
    # initial logging is to memory until we read the config
    logger_startup()

    # parse arguments
    args = parse_args()

    exit_code = compat.asyncio.run(start_server(config=args.config))
    sys.exit(exit_code)



import argparse

from knot_resolver.client.command import CommandArgs

KRES_CLIENT_NAME = "kresctl"


class KresClient:
    def __init__(
        self,
        namespace: argparse.Namespace,
        parser: argparse.ArgumentParser,
        prompt: str = KRES_CLIENT_NAME,
    ) -> None:
        self.path = None
        self.prompt = prompt
        self.namespace = namespace
        self.parser = parser

    def execute(self) -> None:
        if hasattr(self.namespace, "command"):
            args = CommandArgs(self.namespace, self.parser)
            command = args.command(self.namespace)
            command.run(args)
        else:
            self.parser.print_help()

    def _prompt_format(self) -> str:
        bolt = "\033[1m"
        white = "\033[38;5;255m"
        reset = "\033[0;0m"

        prompt = f"{bolt}[{self.prompt} {white}{self.path}{reset}{bolt}]" if self.path else f"{bolt}{self.prompt}"
        return f"{prompt}> {reset}"

    def interactive(self) -> None:
        try:
            while True:
                pass
                # TODO: not working yet
                # cmd = input(f"{self._prompt_format()}")
                # namespace = self.parser.parse_args(cmd.split(" "))
                # namespace.interactive = True
                # namespace.socket = self.namespace.socket
                # self.namespace = namespace
                # self.execute()
        except KeyboardInterrupt:
            pass



# ruff: noqa: D205, D400, D415
import logging
import os
import stat
from enum import Flag, auto
from grp import getgrnam
from pathlib import Path
from pwd import getpwnam, getpwuid
from typing import Any, Dict, Tuple, Type, TypeVar

from knot_resolver.constants import GROUP, USER
from knot_resolver.datamodel.globals import get_permissions_default, get_resolve_root, get_strict_validation
from knot_resolver.utils.modeling.base_value_type import BaseValueType

logger = logging.getLogger(__name__)


class UncheckedPath(BaseValueType):
    """
    Wrapper around pathlib.Path object.

    Can represent pretty much any Path. No checks are performed on the value. The value is taken as is.
    """

    _value: Path

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        self._object_path: str = object_path
        self._parents: Tuple[UncheckedPath, ...] = parents
        self.strict_validation: bool = get_strict_validation()

        if isinstance(source_value, str):
            # we do not load global validation context if the path is absolute
            # this prevents errors when constructing defaults in the schema
            resolve_root = Path("/") if source_value.startswith("/") else get_resolve_root()

            self._raw_value: str = source_value
            if self._parents:
                pp = map(lambda p: p.to_path(), self._parents)
                self._value: Path = Path(resolve_root, *pp, source_value)
            else:
                self._value: Path = Path(resolve_root, source_value)
        else:
            raise ValueError(f"expected file path in a string, got '{source_value}' with type '{type(source_value)}'.")

    def __str__(self) -> str:
        return str(self._value)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, UncheckedPath):
            return False

        return o._value == self._value

    def __int__(self) -> int:
        raise RuntimeError("Path cannot be converted to type <int>")

    def to_path(self) -> Path:
        return self._value

    def serialize(self) -> Any:
        return self._raw_value

    def relative_to(self, parent: "UncheckedPath") -> "UncheckedPath":
        """Return a path with an added parent part."""
        return UncheckedPath(self._raw_value, parents=(parent, *self._parents), object_path=self._object_path)

    UPT = TypeVar("UPT", bound="UncheckedPath")

    def reconstruct(self, cls: Type[UPT]) -> UPT:
        """Rebuild this object as an instance of its subclass. Practically, allows for conversions from."""
        return cls(self._raw_value, parents=self._parents, object_path=self._object_path)

    @classmethod
    def json_schema(cls: Type["UncheckedPath"]) -> Dict[Any, Any]:
        return {
            "type": "string",
        }


class Dir(UncheckedPath):
    """
    Path, that is enforced to be:
    - an existing directory
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        try:
            super().__init__(source_value, parents=parents, object_path=object_path)
            if self.strict_validation and not self._value.is_dir():
                raise ValueError(f"path '{self._value}' does not point to an existing directory")
        except PermissionError as e:
            raise ValueError(str(e)) from e


class AbsoluteDir(Dir):
    """
    Path, that is enforced to be:
    - absolute
    - an existing directory
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        super().__init__(source_value, parents=parents, object_path=object_path)
        if self.strict_validation and not self._value.is_absolute():
            raise ValueError(f"path '{self._value}' is not absolute")


class File(UncheckedPath):
    """
    Path, that is enforced to be:
    - an existing file
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        try:
            super().__init__(source_value, parents=parents, object_path=object_path)
            if self.strict_validation and not self._value.exists():
                raise ValueError(f"file '{self._value}' does not exist")
            if self.strict_validation and not self._value.is_file():
                raise ValueError(f"path '{self._value}' is not a file")
        except PermissionError as e:
            raise ValueError(str(e)) from e


class FilePath(UncheckedPath):
    """
    Path, that is enforced to be:
    - parent of the last path segment is an existing directory
    - it does not point to a dir
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        try:
            super().__init__(source_value, parents=parents, object_path=object_path)
            p = self._value.parent
            if self.strict_validation and (not p.exists() or not p.is_dir()):
                raise ValueError(f"path '{self._value}' does not point inside an existing directory")
            if self.strict_validation and self._value.is_dir():
                raise ValueError(f"path '{self._value}' points to a directory when we expected a file")
        except PermissionError as e:
            raise ValueError(str(e)) from e


class _PermissionMode(Flag):
    READ = auto()
    WRITE = auto()
    EXECUTE = auto()


def _check_permission(dest_path: Path, perm_mode: _PermissionMode) -> bool:
    chflags = {
        _PermissionMode.READ: [stat.S_IRUSR, stat.S_IRGRP, stat.S_IROTH],
        _PermissionMode.WRITE: [stat.S_IWUSR, stat.S_IWGRP, stat.S_IWOTH],
        _PermissionMode.EXECUTE: [stat.S_IXUSR, stat.S_IXGRP, stat.S_IXOTH],
    }

    # running outside the manager (client, ...)
    if get_permissions_default():
        user_uid = getpwnam(USER).pw_uid
        user_gid = getgrnam(GROUP).gr_gid
        username = USER
    # running under root privileges
    elif os.geteuid() == 0:
        return True
    # running normally under an unprivileged user
    else:
        user_uid = os.getuid()
        user_gid = os.getgid()
        username = getpwuid(user_uid).pw_name

    dest_stat = os.stat(dest_path)
    dest_uid = dest_stat.st_uid
    dest_gid = dest_stat.st_gid
    dest_mode = dest_stat.st_mode

    def accessible(perm: _PermissionMode) -> bool:
        if user_uid == dest_uid:
            return bool(dest_mode & chflags[perm][0])
        b_groups = os.getgrouplist(username, user_gid)
        if user_gid == dest_gid or dest_gid in b_groups:
            return bool(dest_mode & chflags[perm][1])
        return bool(dest_mode & chflags[perm][2])

    # __iter__ for class enum.Flag added in python3.11
    # 'for perm in perm_mode:' fails for <=python3.11
    return all(not (perm in perm_mode and not accessible(perm)) for perm in _PermissionMode)


class ReadableFile(File):
    """
    Path, that is enforced to be:

    - an existing file
    - readable by knot-resolver processes
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        super().__init__(source_value, parents=parents, object_path=object_path)

        if self.strict_validation and not _check_permission(self._value, _PermissionMode.READ):
            msg = f"{USER}:{GROUP} has insufficient permissions to read '{self._value}'"
            if not os.access(self._value, os.R_OK):
                raise ValueError(msg)
            logger.info(f"{msg}, but the resolver can somehow (ACLs, ...) read the file")


class WritableDir(Dir):
    """
    Path, that is enforced to be:
    - an existing directory
    - writable/executable by knot-resolver processes
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        super().__init__(source_value, parents=parents, object_path=object_path)

        if self.strict_validation and not _check_permission(
            self._value, _PermissionMode.WRITE | _PermissionMode.EXECUTE
        ):
            msg = f"{USER}:{GROUP} has insufficient permissions to write/execute '{self._value}'"
            if not os.access(self._value, os.W_OK | os.X_OK):
                raise ValueError(msg)
            logger.info(f"{msg}, but the resolver can somehow (ACLs, ...) write to the directory")


class WritableFilePath(FilePath):
    """
    Path, that is enforced to be:
    - parent of the last path segment is an existing directory
    - it does not point to a dir
    - writable/executable parent directory by knot-resolver processes
    """

    def __init__(
        self, source_value: Any, parents: Tuple["UncheckedPath", ...] = tuple(), object_path: str = "/"
    ) -> None:
        super().__init__(source_value, parents=parents, object_path=object_path)

        if self.strict_validation:
            # check that parent dir is writable
            if not _check_permission(self._value.parent, _PermissionMode.WRITE | _PermissionMode.EXECUTE):
                msg = f"{USER}:{GROUP} has insufficient permissions to write/execute '{self._value.parent}'"
                # os.access() on the dir just provides a more precise message,
                # as the os.access() on the file below check everything in one go
                if not os.access(self._value.parent, os.W_OK | os.X_OK):
                    raise ValueError(msg)
                logger.info(f"{msg}, but the resolver can somehow (ACLs, ...) write to the directory")

            # check that existing file is writable
            if self._value.exists() and not _check_permission(self._value, _PermissionMode.WRITE):
                msg = f"{USER}:{GROUP} has insufficient permissions to write/execute '{self._value}'"
                if not os.access(self._value, os.W_OK):
                    raise ValueError(msg)
                logger.info(f"{msg}, but the resolver can somehow (ACLs, ...) write to the file")



# pylint: disable=comparison-with-callable


import enum
import inspect
import sys
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, TypeVar, Union

from .base_generic_type_wrapper import BaseGenericTypeWrapper

NoneType = type(None)


def get_annotations(obj: Any) -> Dict[str, Any]:
    if hasattr(inspect, "get_annotations"):
        return inspect.get_annotations(obj)
    # TODO(bump to py3.10): Safe to remove. This fallback exists for older versions
    return obj.__dict__.get("__annotations__", {})


def is_optional(tp: Any) -> bool:
    origin = getattr(tp, "__origin__", None)
    args = get_generic_type_arguments(tp)

    return origin == Union and len(args) == 2 and args[1] == NoneType


def is_dict(tp: Any) -> bool:
    return getattr(tp, "__origin__", None) in (Dict, dict)


def is_enum(tp: Any) -> bool:
    return inspect.isclass(tp) and issubclass(tp, enum.Enum)


def is_list(tp: Any) -> bool:
    return getattr(tp, "__origin__", None) in (List, list)


def is_tuple(tp: Any) -> bool:
    return getattr(tp, "__origin__", None) in (Tuple, tuple)


def is_union(tp: Any) -> bool:
    """Return true even for optional types, because they are just a Union[T, NoneType]."""
    return getattr(tp, "__origin__", None) == Union


def is_literal(tp: Any) -> bool:
    if sys.version_info.minor == 6:
        return isinstance(tp, type(Literal))
    return getattr(tp, "__origin__", None) == Literal


def is_generic_type_wrapper(tp: Any) -> bool:
    orig = getattr(tp, "__origin__", None)
    return inspect.isclass(orig) and issubclass(orig, BaseGenericTypeWrapper)


def get_generic_type_arguments(tp: Any) -> List[Any]:
    default: List[Any] = []
    if sys.version_info.minor == 6 and is_literal(tp):
        return getattr(tp, "__values__")
    return getattr(tp, "__args__", default)


def get_generic_type_argument(tp: Any) -> Any:
    """Same as function get_generic_type_arguments, but expects just one type argument."""  # noqa: D401
    args = get_generic_type_arguments(tp)
    assert len(args) == 1
    return args[0]


def get_generic_type_wrapper_argument(tp: Type["BaseGenericTypeWrapper[Any]"]) -> Any:
    assert hasattr(tp, "__origin__")
    origin = getattr(tp, "__origin__")

    assert hasattr(origin, "__orig_bases__")
    orig_base: List[Any] = getattr(origin, "__orig_bases__", [])[0]

    arg = get_generic_type_argument(tp)
    return get_generic_type_argument(orig_base[arg])


def is_none_type(tp: Any) -> bool:
    return tp is None or tp == NoneType


def get_attr_type(obj: Any, attr_name: str) -> Any:
    assert hasattr(obj, attr_name)
    assert hasattr(obj, "__annotations__")
    annot = get_annotations(type(obj))
    assert attr_name in annot
    return annot[attr_name]


T = TypeVar("T")


def get_optional_inner_type(optional: Type[Optional[T]]) -> Type[T]:
    assert is_optional(optional)
    t: Type[T] = get_generic_type_arguments(optional)[0]
    return t


def is_internal_field_name(field_name: str) -> bool:
    return field_name.startswith("_")



from typing import Any, Callable, Optional, Type, TypeVar

T = TypeVar("T")


def ignore_exceptions_optional(
    _tp: Type[T], default: Optional[T], *exceptions: Type[BaseException]
) -> Callable[[Callable[..., Optional[T]]], Callable[..., Optional[T]]]:
    """
    Wrap function preventing it from raising exceptions and instead returning the configured default value.

    :param type[T] _tp: Return type of the function. Essentialy only a template argument for type-checking
    :param T default: The value to return as a default
    :param list[Type[BaseException]] exceptions: The list of exceptions to catch
    :return: value of the decorated function, or default if exception raised
    :rtype: T
    """

    def decorator(func: Callable[..., Optional[T]]) -> Callable[..., Optional[T]]:
        def f(*nargs: Any, **nkwargs: Any) -> Optional[T]:
            try:
                return func(*nargs, **nkwargs)
            except BaseException as e:
                if isinstance(e, exceptions):
                    return default
                raise

        return f

    return decorator


def ignore_exceptions(
    default: T, *exceptions: Type[BaseException]
) -> Callable[[Callable[..., Optional[T]]], Callable[..., Optional[T]]]:
    return ignore_exceptions_optional(type(default), default, *exceptions)


def phantom_use(var: Any) -> None:  # pylint: disable=unused-argument
    """
    Consumes argument doing absolutely nothing with it.

    Useful for convincing pylint, that we need the variable even when its unused.
    """



import os
from typing import Any, List, Literal, Optional, Set, Type, Union, cast

from knot_resolver.datamodel.types import WritableFilePath
from knot_resolver.utils.modeling import ConfigSchema
from knot_resolver.utils.modeling.base_schema import is_obj_type_valid

LogLevelEnum = Literal["crit", "err", "warning", "notice", "info", "debug"]
LogTargetEnum = Literal["syslog", "stderr", "stdout"]

LogGroupsProcessesEnum = Literal[
    "manager",
    "supervisord",
    "policy-loader",
    "kresd",
    "cache-gc",
]

LogGroupsManagerEnum = Literal[
    "files",
    "metrics",
    "server",
]

LogGroupsKresdEnum = Literal[
    ## Now the LOG_GRP_*_TAG defines, exactly from ../../../lib/log.h
    "system",
    "cache",
    "io",
    "net",
    "ta",
    "tasent",
    "tasign",
    "taupd",
    "tls",
    "gnutls",
    "tls_cl",
    "xdp",
    "doh",
    "dnssec",
    "hint",
    "plan",
    "iterat",
    "valdtr",
    "resolv",
    "select",
    "zoncut",
    "cookie",
    "statis",
    "rebind",
    "worker",
    "policy",
    "daf",
    "timejm",
    "timesk",
    "graphi",
    "prefil",
    "primin",
    "srvstl",
    "wtchdg",
    "nsid",
    "dnstap",
    "tests",
    "dotaut",
    "http",
    "contrl",
    "module",
    "devel",
    "renum",
    "exterr",
    "rules",
    "prlayr",
    "defer",
    "doq",
    "ngtcp2",
    # "reqdbg",... (non-displayed section of the enum)
]

LogGroupsEnum = Literal[LogGroupsProcessesEnum, LogGroupsManagerEnum, LogGroupsKresdEnum]


class DnstapSchema(ConfigSchema):
    """
    Logging DNS queries and responses to a unix socket.

    ---
    enable: Enable/disable DNS queries logging.
    unix_socket: Path to unix domain socket where dnstap messages will be sent.
    log_queries: Log queries from downstream in wire format.
    log_responses: Log responses to downstream in wire format.
    log_tcp_rtt: Log TCP RTT (Round-trip time).
    """

    enable: bool = False
    unix_socket: Optional[WritableFilePath] = None
    log_queries: bool = False
    log_responses: bool = False
    log_tcp_rtt: bool = False

    def _validate(self) -> None:
        if self.enable and self.unix_socket is None:
            raise ValueError("DNS queries logging enabled, but 'unix-socket' not specified")


class LoggingSchema(ConfigSchema):
    class Raw(ConfigSchema):
        """
        Logging and debugging configuration.

        ---
        level: Global logging level.
        target: Global logging stream target. "from-env" uses $KRES_LOGGING_TARGET and defaults to "stdout".
        groups: List of groups for which 'debug' logging level is set.
        dnstap: Logging DNS requests and responses to a unix socket.
        """

        level: LogLevelEnum = "notice"
        target: Union[LogTargetEnum, Literal["from-env"]] = "from-env"
        groups: Optional[List[LogGroupsEnum]] = None
        dnstap: DnstapSchema = DnstapSchema()

    _LAYER = Raw

    level: LogLevelEnum
    target: LogTargetEnum
    groups: Optional[List[LogGroupsEnum]]
    dnstap: DnstapSchema

    def _target(self, raw: Raw) -> LogTargetEnum:
        if raw.target == "from-env":
            target = os.environ.get("KRES_LOGGING_TARGET") or "stdout"
            if not is_obj_type_valid(target, cast(Type[Any], LogTargetEnum)):
                raise ValueError(f"logging target '{target}' read from $KRES_LOGGING_TARGET is invalid")
            return cast(LogTargetEnum, target)
        return raw.target

    def _validate(self) -> None:
        if self.groups is None:
            return

        checked: Set[str] = set()
        for i, g in enumerate(self.groups):
            if g in checked:
                raise ValueError(f"duplicate logging group '{g}' on index {i}")
            checked.add(g)



import asyncio
from asyncio import Lock
from typing import Any, Awaitable, Callable, List, Tuple

from knot_resolver.datamodel import KresConfig
from knot_resolver.utils.functional import Result
from knot_resolver.utils.modeling.exceptions import DataParsingError
from knot_resolver.utils.modeling.types import NoneType

from .exceptions import KresManagerBaseError

VerifyCallback = Callable[[KresConfig, KresConfig, bool], Awaitable[Result[None, str]]]
UpdateCallback = Callable[[KresConfig, bool], Awaitable[None]]


class ConfigStore:
    def __init__(self, initial_config: KresConfig) -> None:
        self._config = initial_config
        self._verifiers: List[VerifyCallback] = []
        self._callbacks: List[UpdateCallback] = []
        self._update_lock: Lock = Lock()

    async def update(self, config: KresConfig, force: bool = False) -> None:
        # invoke pre-change verifiers
        results: Tuple[Result[None, str], ...] = tuple(
            await asyncio.gather(*[ver(self._config, config, force) for ver in self._verifiers])
        )
        err_res = filter(lambda r: r.is_err(), results)
        errs = list(map(lambda r: r.unwrap_err(), err_res))
        if len(errs) > 0:
            raise KresManagerBaseError("Configuration validation failed. The reasons are:\n - " + "\n - ".join(errs))

        async with self._update_lock:
            # update the stored config with the new version
            self._config = config

            # invoke change callbacks
            for call in self._callbacks:
                await call(config, force)

    async def renew(self, force: bool = False) -> None:
        await self.update(self._config, force)

    async def register_verifier(self, verifier: VerifyCallback) -> None:
        self._verifiers.append(verifier)
        res = await verifier(self.get(), self.get(), False)
        if res.is_err():
            raise DataParsingError(f"Initial config verification failed with error: {res.unwrap_err()}")

    async def register_on_change_callback(self, callback: UpdateCallback) -> None:
        """Register new callback and immediately call it with current config."""
        self._callbacks.append(callback)
        await callback(self.get(), False)

    def get(self) -> KresConfig:
        return self._config


def only_on_no_changes_update(selector: Callable[[KresConfig], Any]) -> Callable[[UpdateCallback], UpdateCallback]:
    def decorator(orig_func: UpdateCallback) -> UpdateCallback:
        original_value_set: Any = False
        original_value: Any = None

        async def new_func_update(config: KresConfig, force: bool = False) -> None:
            nonlocal original_value_set
            nonlocal original_value
            if not original_value_set:
                original_value_set = True
            elif original_value == selector(config) or force:
                await orig_func(config, force)
            original_value = selector(config)

        return new_func_update

    return decorator


def only_on_real_changes_update(selector: Callable[[KresConfig], Any]) -> Callable[[UpdateCallback], UpdateCallback]:
    def decorator(orig_func: UpdateCallback) -> UpdateCallback:
        original_value_set: Any = False
        original_value: Any = None

        async def new_func_update(config: KresConfig, force: bool) -> None:
            nonlocal original_value_set
            nonlocal original_value
            if not original_value_set:
                original_value_set = True
                await orig_func(config, force)
            elif original_value != selector(config) or force:
                await orig_func(config, force)
            original_value = selector(config)

        return new_func_update

    return decorator


def only_on_real_changes_verifier(selector: Callable[[KresConfig], Any]) -> Callable[[VerifyCallback], VerifyCallback]:
    def decorator(orig_func: VerifyCallback) -> VerifyCallback:
        original_value_set: Any = False
        original_value: Any = None

        async def new_func_verifier(old: KresConfig, new: KresConfig, force: bool) -> Result[NoneType, str]:
            nonlocal original_value_set
            nonlocal original_value
            if not original_value_set:
                original_value_set = True
                original_value = selector(new)
                await orig_func(old, new, force)
            elif original_value != selector(new):
                original_value = selector(new)
                await orig_func(old, new, force)
            return Result.ok(None)

        return new_func_verifier

    return decorator



from typing import Optional

from knot_resolver.datamodel.types import (
    Int0_32,
    IntPositive,
    TimeUnit,
)
from knot_resolver.utils.modeling import ConfigSchema


class RateLimitingSchema(ConfigSchema):
    """
    Configuration of rate limiting.

    ---
    enable: Enable/disable rate limiting
    rate_limit: Maximal number of allowed queries per second from a single host.
    instant_limit: Maximal number of allowed queries at a single point in time from a single host.
    capacity: Expected maximal number of blocked networks/hosts at the same time.
    slip: Number of restricted responses out of which one is sent as truncated, the others are dropped.
    log_period: Minimal time between two log messages, or '0s' to disable.
    dry_run: Perform only classification and logging but no restrictions.
    """

    enable: bool = False
    rate_limit: Optional[IntPositive] = None
    instant_limit: IntPositive = IntPositive(50)
    capacity: IntPositive = IntPositive(524288)
    slip: Int0_32 = Int0_32(2)
    log_period: TimeUnit = TimeUnit("0s")
    dry_run: bool = False

    def _validate(self) -> None:
        if self.enable and not self.rate_limit:
            raise ValueError("'rate-limit' has to be configured to enable rate limiting")

        max_instant_limit = int(2**32 // 768 - 1)
        if not int(self.instant_limit) <= max_instant_limit:
            raise ValueError(f"'instant-limit' has to be in range 1..{max_instant_limit}")
        if self.rate_limit and not int(self.rate_limit) <= 1000 * int(self.instant_limit):
            raise ValueError("'rate-limit' has to be in range 1..(1000 * instant-limit)")



import importlib.util
import platform
import re
from pathlib import Path

VERSION = "6.2.0"
USER = "knot-resolver"
GROUP = "knot-resolver"

# dirs paths
RUN_DIR = Path("/run/knot-resolver")
ETC_DIR = Path("/etc/knot-resolver")
SBIN_DIR = Path("/usr/bin")
CACHE_DIR = Path("/var/cache/knot-resolver")

# files paths
CONFIG_FILE = ETC_DIR / "config.yaml"
API_SOCK_FILE = RUN_DIR / "kres-api.sock"

# executables paths
KRESD_EXECUTABLE = SBIN_DIR / "kresd"
KRES_CACHE_GC_EXECUTABLE = SBIN_DIR / "kres-cache-gc"

LINUX_SYS = platform.system() == "Linux"
FREEBSD_SYS = platform.system() == "FreeBSD"

WATCHDOG_LIB = bool(importlib.util.find_spec("watchdog"))
PROMETHEUS_LIB = bool(importlib.util.find_spec("prometheus_client"))


def _freebsd_workers_support() -> bool:
    if FREEBSD_SYS:
        release = platform.release()
        match = re.match(r"(\d+)", release)
        if match:
            freebsd_min_version = 12
            return int(match.group(1)) >= freebsd_min_version
    return False


# It is possible to configure multiple kresd workers on Linux systems due to the SO_REUSEPORT socket option.
# FreeBSD version >=12 supports it specifically due to the additional SO_REUSEPORT_LB socket option.
WORKERS_SUPPORT = LINUX_SYS or _freebsd_workers_support()

# Systemd-like NOTIFY message is supported only on Linux systems
NOTIFY_SUPPORT = LINUX_SYS



# type: ignore
# ruff: noqa: SLF001
# pylint: disable=c-extension-no-member

from knot_resolver.constants import NOTIFY_SUPPORT

if NOTIFY_SUPPORT:
    import os
    import signal
    import time
    from functools import partial
    from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

    from supervisor.events import ProcessStateEvent, ProcessStateStartingEvent, subscribe
    from supervisor.medusa.asyncore_25 import compact_traceback
    from supervisor.process import Subprocess
    from supervisor.states import ProcessStates
    from supervisor.supervisord import Supervisor

    from knot_resolver.controller.supervisord.plugin import notify

    starting_processes: List[Subprocess] = []

    def is_type_notify(proc: Subprocess) -> bool:
        return (
            proc.config.environment is not None and proc.config.environment.get("X-SUPERVISORD-TYPE", None) == "notify"
        )

    class NotifySocketDispatcher:
        """
        See supervisor.dispatcher
        """

        def __init__(self, supervisor: Supervisor, fd: int):
            self._supervisor = supervisor
            self.fd = fd
            self.closed = False  # True if close() has been called

        def __repr__(self):
            return f"<{self.__class__.__name__} with fd={self.fd}>"

        def readable(self):
            return True

        def writable(self):
            return False

        def handle_read_event(self):
            logger: Any = self._supervisor.options.logger

            res: Optional[Tuple[int, bytes]] = notify.read_message(self.fd)
            if res is None:
                return  # there was some junk
            pid, data = res

            # pylint: disable=undefined-loop-variable
            for proc in starting_processes:
                if proc.pid == pid:
                    break
            else:
                logger.warn(f"ignoring ready notification from unregistered PID={pid}")
                return

            if data.startswith(b"READY=1"):
                # handle case, when some process is really ready

                if is_type_notify(proc):
                    proc._assertInState(ProcessStates.STARTING)
                    proc.change_state(ProcessStates.RUNNING)
                    logger.info(
                        f"success: {proc.config.name} entered RUNNING state, process sent notification via $NOTIFY_SOCKET"
                    )
                else:
                    logger.warn(
                        f"ignoring READY notification from {proc.config.name}, which is not configured to send it"
                    )

            elif data.startswith(b"STOPPING=1"):
                # just accept the message, filter unwanted notifications and do nothing else

                if is_type_notify(proc):
                    logger.info(
                        f"success: {proc.config.name} entered STOPPING state, process sent notification via $NOTIFY_SOCKET"
                    )
                else:
                    logger.warn(
                        f"ignoring STOPPING notification from {proc.config.name}, which is not configured to send it"
                    )

            else:
                # handle case, when we got something unexpected
                logger.warn(f"ignoring unrecognized data on $NOTIFY_SOCKET sent from PID={pid}, data='{data!r}'")
                return

        def handle_write_event(self):
            raise ValueError("this dispatcher is not writable")

        def handle_error(self):
            _nil, t, v, tbinfo = compact_traceback()

            self._supervisor.options.logger.error(
                f"uncaptured python exception, closing notify socket {repr(self)} ({t}:{v} {tbinfo})"
            )
            self.close()

        def close(self):
            if not self.closed:
                os.close(self.fd)
                self.closed = True

        def flush(self):
            pass

    def keep_track_of_starting_processes(event: ProcessStateEvent) -> None:
        global starting_processes

        proc: Subprocess = event.process

        if isinstance(event, ProcessStateStartingEvent):
            # process is starting
            # if proc not in starting_processes:
            starting_processes.append(proc)

        else:
            # not starting
            starting_processes = [p for p in starting_processes if p.pid is not proc.pid]

    notify_dispatcher: Optional[NotifySocketDispatcher] = None

    def process_transition(slf: Subprocess) -> None:
        if not is_type_notify(slf):
            return slf

        # modified version of upstream process transition code
        if slf.state == ProcessStates.STARTING:
            if time.time() - slf.laststart > slf.config.startsecs:
                # STARTING -> STOPPING if the process has not sent ready notification
                # within proc.config.startsecs
                slf.config.options.logger.warn(
                    f"process '{slf.config.name}' did not send ready notification within {slf.config.startsecs} secs, killing"
                )
                slf.kill(signal.SIGKILL)
                slf.x_notifykilled = True  # used in finish() function to set to FATAL state
                slf.laststart = time.time() + 1  # prevent immediate state transition to RUNNING from happening

        # return self for chaining
        return slf

    def subprocess_finish_tail(slf, pid, sts) -> Tuple[Any, Any, Any]:
        if getattr(slf, "x_notifykilled", False):
            # we want FATAL, not STOPPED state after timeout waiting for startup notification
            # why? because it's likely not gonna help to try starting the process up again if
            # it failed so early
            slf.change_state(ProcessStates.FATAL)

            # clear the marker value
            del slf.x_notifykilled

        # return for chaining
        return slf, pid, sts

    def supervisord_get_process_map(supervisord: Any, mp: Dict[Any, Any]) -> Dict[Any, Any]:
        global notify_dispatcher
        if notify_dispatcher is None:
            notify_dispatcher = NotifySocketDispatcher(supervisord, notify.init_socket())
            supervisord.options.logger.info("notify: injected $NOTIFY_SOCKET into event loop")

        # add our dispatcher to the result
        assert notify_dispatcher.fd not in mp
        mp[notify_dispatcher.fd] = notify_dispatcher

        return mp

    def process_spawn_as_child_add_env(slf: Subprocess, *args: Any) -> Tuple[Any, ...]:
        if is_type_notify(slf):
            slf.config.environment["NOTIFY_SOCKET"] = os.getcwd() + "/supervisor-notify-socket"
        return (slf, *args)

    T = TypeVar("T")
    U = TypeVar("U")

    def chain(first: Callable[..., U], second: Callable[[U], T]) -> Callable[..., T]:
        def wrapper(*args: Any, **kwargs: Any) -> T:
            res = first(*args, **kwargs)
            if isinstance(res, tuple):
                return second(*res)
            return second(res)

        return wrapper

    def append(first: Callable[..., T], second: Callable[..., None]) -> Callable[..., T]:
        def wrapper(*args: Any, **kwargs: Any) -> T:
            res = first(*args, **kwargs)
            second(*args, **kwargs)
            return res

        return wrapper

    def monkeypatch(supervisord: Supervisor) -> None:
        """Inject ourselves into supervisord code"""

        # append notify socket handler to event loop
        supervisord.get_process_map = chain(
            supervisord.get_process_map, partial(supervisord_get_process_map, supervisord)
        )

        # prepend timeout handler to transition method
        Subprocess.transition = chain(process_transition, Subprocess.transition)
        Subprocess.finish = append(Subprocess.finish, subprocess_finish_tail)

        # add environment variable $NOTIFY_SOCKET to starting processes
        Subprocess._spawn_as_child = chain(process_spawn_as_child_add_env, Subprocess._spawn_as_child)

        # keep references to starting subprocesses
        subscribe(ProcessStateEvent, keep_track_of_starting_processes)

    def inject(supervisord: Supervisor, **_config: Any) -> Any:  # pylint: disable=useless-return
        monkeypatch(supervisord)

        # this method is called by supervisord when loading the plugin,
        # it should return XML-RPC object, which we don't care about
        # That's why why are returning just None
        return None



import copy
from abc import ABC, abstractmethod  # pylint: disable=[no-name-in-module]
from typing import Any, List, Literal, Optional, Tuple, Union

from knot_resolver.utils.modeling.base_schema import BaseSchema, map_object
from knot_resolver.utils.modeling.json_pointer import json_ptr_resolve


class PatchError(Exception):
    pass


class Op(BaseSchema, ABC):
    @abstractmethod
    def eval(self, fakeroot: Any) -> Any:
        """Modify the given fakeroot, returns a new one."""

    def _resolve_ptr(self, fakeroot: Any, ptr: str) -> Tuple[Any, Any, Union[str, int, None]]:
        # Lookup tree part based on the given JSON pointer
        parent, obj, token = json_ptr_resolve(fakeroot["root"], ptr)

        # the lookup was on pure data, wrap the results in QueryTree
        if parent is None:
            parent = fakeroot
            token = "root"

        assert token is not None

        return parent, obj, token


class AddOp(Op):
    op: Literal["add"]
    path: str
    value: Any

    def eval(self, fakeroot: Any) -> Any:
        parent, _obj, token = self._resolve_ptr(fakeroot, self.path)

        if isinstance(parent, dict):
            parent[token] = self.value
        elif isinstance(parent, list):
            if token == "-":
                parent.append(self.value)
            else:
                assert isinstance(token, int)
                parent.insert(token, self.value)
        else:
            raise AssertionError("never happens")

        return fakeroot


class RemoveOp(Op):
    op: Literal["remove"]
    path: str

    def eval(self, fakeroot: Any) -> Any:
        parent, _obj, token = self._resolve_ptr(fakeroot, self.path)
        del parent[token]
        return fakeroot


class ReplaceOp(Op):
    op: Literal["replace"]
    path: str
    value: str

    def eval(self, fakeroot: Any) -> Any:
        parent, obj, token = self._resolve_ptr(fakeroot, self.path)

        if obj is None:
            raise PatchError("the value you are trying to replace is null")
        parent[token] = self.value
        return fakeroot


class MoveOp(Op):
    op: Literal["move"]
    source: str
    path: str

    def _source(self, source: Any) -> Any:
        if "from" not in source:
            raise ValueError("missing property 'from' in 'move' JSON patch operation")
        return str(source["from"])

    def eval(self, fakeroot: Any) -> Any:
        if self.path.startswith(self.source):
            raise PatchError("can't move value into itself")

        _parent, obj, _token = self._resolve_ptr(fakeroot, self.source)
        newobj = copy.deepcopy(obj)

        fakeroot = RemoveOp({"op": "remove", "path": self.source}).eval(fakeroot)
        return AddOp({"path": self.path, "value": newobj, "op": "add"}).eval(fakeroot)


class CopyOp(Op):
    op: Literal["copy"]
    source: str
    path: str

    def _source(self, source: Any) -> Any:
        if "from" not in source:
            raise ValueError("missing property 'from' in 'copy' JSON patch operation")
        return str(source["from"])

    def eval(self, fakeroot: Any) -> Any:
        _parent, obj, _token = self._resolve_ptr(fakeroot, self.source)
        newobj = copy.deepcopy(obj)

        return AddOp({"path": self.path, "value": newobj, "op": "add"}).eval(fakeroot)


class TestOp(Op):
    op: Literal["test"]
    path: str
    value: Any

    def eval(self, fakeroot: Any) -> Any:
        _parent, obj, _token = self._resolve_ptr(fakeroot, self.path)

        if obj != self.value:
            raise PatchError("test failed")

        return fakeroot


def query(
    original: Any, method: Literal["get", "delete", "put", "patch"], ptr: str, payload: Any
) -> Tuple[Any, Optional[Any]]:
    ########################################
    # Prepare data we will be working on

    # First of all, we consider the original data to be immutable. So we need to make a copy
    # in order to freely mutate them
    dataroot = copy.deepcopy(original)

    # To simplify referencing the root, create a fake root node
    fakeroot = {"root": dataroot}

    #########################################
    # Handle the actual requested operation

    # get = return what the path selector picks
    if method == "get":
        parent, obj, token = json_ptr_resolve(fakeroot, f"/root{ptr}")
        return fakeroot["root"], obj

    if method == "delete":
        fakeroot = RemoveOp({"op": "remove", "path": ptr}).eval(fakeroot)
        return fakeroot["root"], None

    if method == "put":
        parent, obj, token = json_ptr_resolve(fakeroot, f"/root{ptr}")
        assert parent is not None  # we know this due to the fakeroot
        if isinstance(parent, list) and token == "-":
            parent.append(payload)
        else:
            parent[token] = payload
        return fakeroot["root"], None

    if method == "patch":
        tp = List[Union[AddOp, RemoveOp, MoveOp, CopyOp, TestOp, ReplaceOp]]
        transaction: tp = map_object(tp, payload)

        for i, op in enumerate(transaction):
            try:
                fakeroot = op.eval(fakeroot)
            except PatchError as e:
                raise ValueError(f"json patch transaction failed on step {i}") from e

        return fakeroot["root"], None

    raise AssertionError("invalid operation, never happens")



# type: ignore
# pylint: disable=protected-access
import atexit
import os
import signal
from typing import Any, Optional

from supervisor.compat import as_string
from supervisor.events import ProcessStateFatalEvent, ProcessStateRunningEvent, ProcessStateStartingEvent, subscribe
from supervisor.options import ServerOptions
from supervisor.process import Subprocess
from supervisor.states import SupervisorStates
from supervisor.supervisord import Supervisor

from knot_resolver.utils.systemd_notify import systemd_notify

superd: Optional[Supervisor] = None


def check_for_fatal_manager(event: ProcessStateFatalEvent) -> None:
    assert superd is not None

    proc: Subprocess = event.process
    processname = as_string(proc.config.name)
    if processname == "manager":
        # stop the whole supervisord gracefully
        superd.options.logger.critical("manager process entered FATAL state! Shutting down")
        superd.options.mood = SupervisorStates.SHUTDOWN

        # force the interpreter to exit with exit code 1
        atexit.register(lambda: os._exit(1))


def check_for_starting_manager(event: ProcessStateStartingEvent) -> None:
    assert superd is not None

    proc: Subprocess = event.process
    processname = as_string(proc.config.name)
    if processname == "manager":
        # manager has sucessfully started, report it upstream
        systemd_notify(STATUS="Starting services...")


def check_for_runnning_manager(event: ProcessStateRunningEvent) -> None:
    assert superd is not None

    proc: Subprocess = event.process
    processname = as_string(proc.config.name)
    if processname == "manager":
        # manager has sucessfully started, report it upstream
        systemd_notify(READY="1", STATUS="Ready")


def get_server_options_signal(self):
    sig = self.signal_receiver.get_signal()
    if sig == signal.SIGHUP and superd is not None:
        superd.options.logger.info("received SIGHUP, forwarding to the process 'manager'")
        manager_pid = superd.process_groups["manager"].processes["manager"].pid
        os.kill(manager_pid, signal.SIGHUP)
        return None

    return sig


def inject(supervisord: Supervisor, **_config: Any) -> Any:  # pylint: disable=useless-return
    global superd
    superd = supervisord

    # This status notification here unsets the env variable $NOTIFY_SOCKET provided by systemd
    # and stores it locally. Therefore, it shouldn't clash with $NOTIFY_SOCKET we are providing
    # downstream
    systemd_notify(STATUS="Initializing supervisord...")

    # register events
    subscribe(ProcessStateFatalEvent, check_for_fatal_manager)
    subscribe(ProcessStateStartingEvent, check_for_starting_manager)
    subscribe(ProcessStateRunningEvent, check_for_runnning_manager)

    # forward SIGHUP to manager
    ServerOptions.get_signal = get_server_options_signal

    # this method is called by supervisord when loading the plugin,
    # it should return XML-RPC object, which we don't care about
    # That's why why are returning just None
    return None



# noqa: INP001
import argparse
import sys
from typing import List, Optional, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, comp_get_words, register_command
from knot_resolver.utils.modeling.parsing import DataFormat, parse_json
from knot_resolver.utils.requests import request


@register_command
class MetricsCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        self.file: Optional[str] = namespace.file
        self.prometheus: bool = namespace.prometheus

        super().__init__(namespace)

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        metrics = subparser.add_parser(
            "metrics",
            help="Get aggregated metrics from the running resolver"
            " in JSON format (default) or optionally in Prometheus format."
            "\nThe 'prometheus-client' Python package needs to be installed if you wish to use the Prometheus format."
            "\nRequires a connection to the management HTTP API.",
        )

        metrics.add_argument(
            "--prometheus",
            help="Get metrics in Prometheus format if dependencies are met in the resolver.",
            action="store_true",
            default=False,
        )

        metrics.add_argument(
            "file",
            help="Optional. The file into which metrics will be exported."
            "\nIf not specified, the metrics are printed into stdout.",
            nargs="?",
            default=None,
        )
        return metrics, MetricsCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:
        response = request(args.socket, "GET", "metrics/prometheus" if self.prometheus else "metrics/json")

        if response.status == 200:
            if self.prometheus:
                metrics = response.body
            else:
                metrics = DataFormat.JSON.dict_dump(parse_json(response.body), indent=4)

            if self.file:
                with open(self.file, "w") as f:
                    f.write(metrics)
            else:
                print(metrics)
        else:
            print(response, file=sys.stderr)
            if self.prometheus and response.status == 404:
                print("Prometheus is unavailable due to missing optional dependencies", file=sys.stderr)
            sys.exit(1)



"""
The parsing and validation of the datamodel is dependent on a global state:
- a file system path used for resolving relative paths.

Commentary from @vsraier:
=========================

While this is not ideal, it is the best we can do at the moment. When I created this module,
the datamodel was dependent on the global state implicitly. The validation procedures just read
the current working directory. This module is the first step in removing the global dependency.

At some point in the future, it might be interesting to add something like a "validation context"
to the modelling tools. It is not technically complicated, but it requires
massive model changes I am not willing to make at the moment. Ideally, when implementing this,
the BaseSchema would turn into an empty class without any logic. Not even a constructor. All logic
would be in the ObjectMapper class. Similar to how Gson works in Java or AutoMapper in C#.
"""  # noqa: D205

from pathlib import Path
from typing import Optional


class Context:
    resolve_root: Optional[Path]
    strict_validation: bool
    permissions_default: bool

    def __init__(
        self, resolve_root: Optional[Path], strict_validation: bool = True, permissions_default: bool = True
    ) -> None:
        self.resolve_root = resolve_root
        self.strict_validation = strict_validation
        self.permissions_default = permissions_default


_global_context: Context = Context(None)


def set_global_validation_context(context: Context) -> None:
    global _global_context
    _global_context = context


def get_global_validation_context() -> Context:
    return _global_context


def reset_global_validation_context() -> None:
    global _global_context
    _global_context = Context(None)


def get_resolve_root() -> Path:
    if _global_context.resolve_root is None:
        raise RuntimeError(
            "Global validation context 'resolve_root' is not set!"
            " Before validation, you have to set it using `set_global_validation_context()` function!"
        )

    return _global_context.resolve_root


def get_strict_validation() -> bool:
    return _global_context.strict_validation


def get_permissions_default() -> bool:
    return _global_context.permissions_default



from typing import List, Literal, Optional

from knot_resolver.datamodel.types import FloatNonNegative, IDPattern, IPNetwork
from knot_resolver.utils.modeling import ConfigSchema


class ViewOptionsSchema(ConfigSchema):
    """
    Configuration options for clients identified by the view.

    ---
    minimize: Send minimum amount of information in recursive queries to enhance privacy.
    dns64: Enable/disable DNS64.
    price_factor: Multiplies rate-limiting and defer prices of operations, use 0 to whitelist.
    fallback: Enable/disable fallback on resolution failure.
    """

    minimize: bool = True
    dns64: bool = True
    price_factor: FloatNonNegative = FloatNonNegative(1.0)
    fallback: bool = True


class ViewSchema(ConfigSchema):
    """
    Configuration parameters that allow you to create personalized policy rules and other.

    ---
    subnets: Identifies the client based on his subnet.  Rule with more precise subnet takes priority.
    dst_subnet: Destination subnet, as an additional condition.
    protocols: Transport protocol, as an additional condition.
    tags: Tags to link with other policy rules.
    answer: Direct approach how to handle request from clients identified by the view.
    options: Configuration options for clients identified by the view.
    """

    subnets: List[IPNetwork]
    dst_subnet: Optional[IPNetwork] = None  # could be a list as well, iterated in template
    protocols: Optional[List[Literal["udp53", "tcp53", "dot", "doh", "doq"]]] = None
    tags: Optional[List[IDPattern]] = None
    answer: Optional[Literal["allow", "refused", "noanswer"]] = None
    options: ViewOptionsSchema = ViewOptionsSchema()

    def _validate(self) -> None:
        if bool(self.tags) == bool(self.answer):
            raise ValueError("exactly one of 'tags' and 'answer' must be configured")



"""
The module contains autodetection logic for available controllers.

Because we have to catch errors from imports, they are located in functions which are invoked at the end of this file.
We supported multiple subprocess controllers while developing it. It now all converged onto just supervisord.
The interface however remains so that different controllers can be added in the future.
"""

import asyncio
import logging
from typing import List, Optional

from knot_resolver.controller.interface import SubprocessController
from knot_resolver.datamodel.config_schema import KresConfig

logger = logging.getLogger(__name__)

"""
List of all subprocess controllers that are available in order of priority.
It is filled dynamically based on available modules that do not fail to import.
"""
_registered_controllers: List[SubprocessController] = []


def try_supervisord() -> None:
    """Attempt to load supervisord controllers."""
    try:
        from knot_resolver.controller.supervisord import SupervisordSubprocessController

        _registered_controllers.append(SupervisordSubprocessController())
    except ImportError:
        logger.error("Failed to import modules related to supervisord service manager", exc_info=True)


async def get_best_controller_implementation(config: KresConfig) -> SubprocessController:
    logger.info("Starting service manager auto-selection...")

    if len(_registered_controllers) == 0:
        logger.error("No controllers are available! Did you install all dependencies?")
        raise LookupError("No service managers available!")

    # check all controllers concurrently
    res = await asyncio.gather(*(cont.is_controller_available(config) for cont in _registered_controllers))
    logger.info(
        "Available subprocess controllers are %s",
        str(tuple((str(c) for r, c in zip(res, _registered_controllers) if r))),
    )

    # take the first one on the list which is available
    for avail, controller in zip(res, _registered_controllers):
        if avail:
            logger.info("Selected controller '%s'", str(controller))
            return controller

    # or fail
    raise LookupError("Can't find any available service manager!")


def list_controller_names() -> List[str]:
    """
    Return a list of names of registered controllers.

    The listed controllers are not necessary functional.
    """
    return [str(controller) for controller in sorted(_registered_controllers, key=str)]


async def get_controller_by_name(config: KresConfig, name: str) -> SubprocessController:
    logger.debug("Subprocess controller selected manualy by the user, testing feasibility...")

    controller: Optional[SubprocessController] = None
    for c in sorted(_registered_controllers, key=str):
        if str(c).startswith(name):
            if str(c) != name:
                logger.debug("Assuming '%s' is a shortcut for '%s'", name, str(c))
            controller = c
            break

    if controller is None:
        logger.error("Subprocess controller with name '%s' was not found", name)
        raise LookupError(f"No subprocess controller named '{name}' found")

    if await controller.is_controller_available(config):
        logger.info("Selected controller '%s'", str(controller))
        return controller
    raise LookupError("The selected subprocess controller is not available for use on this system.")


# run the imports on module load
try_supervisord()



# ruff: noqa: E501
from typing import List, Optional

from knot_resolver.datamodel.types import DomainName, EscapedStr, ReadableFile
from knot_resolver.utils.modeling import ConfigSchema


class TrustAnchorFileSchema(ConfigSchema):
    """
    Trust-anchor zonefile configuration.

    ---
    file: Path to the zonefile that stores trust-anchors.
    read_only: Blocks zonefile updates according to RFC 5011.

    """

    file: ReadableFile
    read_only: bool = False


class DnssecSchema(ConfigSchema):
    """
    DNSSEC configuration.

    ---
    enable: Enable/disable DNSSEC.
    log_bogus: Enable logging for each DNSSEC validation failure if '/logging/level' is set to at least 'notice'.
    sentinel: Allows users of DNSSEC validating resolver to detect which root keys are configured in resolver's chain of trust. (RFC 8509)
    signal_query: Signaling Trust Anchor Knowledge in DNSSEC Using Key Tag Query, according to (RFC 8145#section-5).
    trust_anchors: List of trust-anchors in DS/DNSKEY records format.
    trust_anchors_files: List of zone-files where trust-anchors are stored.
    trust_anchors: Trust-anchors configuration.
    negative_trust_anchors: List of domain names representing negative trust-anchors. (RFC 7646)
    """

    enable: bool = True
    log_bogus: bool = False
    sentinel: bool = True
    signal_query: bool = True
    trust_anchors: Optional[List[EscapedStr]] = None
    trust_anchors_files: Optional[List[TrustAnchorFileSchema]] = None
    negative_trust_anchors: Optional[List[DomainName]] = None



# type: ignore
# pylint: disable=protected-access

import os
import sys
import traceback
from typing import Any, Literal

from supervisor.dispatchers import POutputDispatcher
from supervisor.loggers import LevelsByName, StreamHandler, SyslogHandler
from supervisor.supervisord import Supervisor

FORWARD_LOG_LEVEL = LevelsByName.CRIT  # to make sure it's always printed


def empty_function(*args, **kwargs):
    pass


FORWARD_MSG_FORMAT: str = "%(name)s[%(pid)d]%(stream)s: %(data)s"


def p_output_dispatcher_log(self: POutputDispatcher, data: bytearray):
    if data:
        # parse the input
        if not isinstance(data, bytes):
            text = data
        else:
            try:
                text = data.decode("utf-8")
            except UnicodeDecodeError:
                text = "Undecodable: %r" % data

        # print line by line prepending correct prefix to match the style
        config = self.process.config
        config.options.logger.handlers = forward_handlers
        for line in text.splitlines():
            stream = ""
            if self.channel == "stderr":
                stream = " (stderr)"
            config.options.logger.log(
                FORWARD_LOG_LEVEL, FORWARD_MSG_FORMAT, name=config.name, stream=stream, data=line, pid=self.process.pid
            )
        config.options.logger.handlers = supervisord_handlers


def _create_handler(fmt, level, target: Literal["stdout", "stderr", "syslog"]) -> StreamHandler:
    if target == "syslog":
        handler = SyslogHandler()
    else:
        handler = StreamHandler(sys.stdout if target == "stdout" else sys.stderr)
        handler.setFormat(fmt)
        handler.setLevel(level)
    return handler


supervisord_handlers = []
forward_handlers = []


def inject(supervisord: Supervisor, **config: Any) -> Any:  # pylint: disable=useless-return
    try:
        # reconfigure log handlers
        supervisord.options.logger.info("reconfiguring log handlers")
        supervisord_handlers.append(
            _create_handler(
                f"%(asctime)s supervisor[{os.getpid()}]: [%(levelname)s] %(message)s\n",
                supervisord.options.loglevel,
                config["target"],
            )
        )
        forward_handlers.append(
            _create_handler("%(asctime)s %(message)s\n", supervisord.options.loglevel, config["target"])
        )
        supervisord.options.logger.handlers = supervisord_handlers

        # replace output handler for subprocesses
        POutputDispatcher._log = p_output_dispatcher_log  # noqa: SLF001

        # we forward stdio in all cases, even when logging to syslog. This should prevent the unforturtunate
        # case of swallowing an error message leaving the users confused. To make the forwarded lines obvious
        # we just prepend a explanatory string at the beginning of all messages
        if config["target"] == "syslog":
            global FORWARD_MSG_FORMAT
            FORWARD_MSG_FORMAT = "captured stdio output from " + FORWARD_MSG_FORMAT

        # this method is called by supervisord when loading the plugin,
        # it should return XML-RPC object, which we don't care about
        # That's why why are returning just None
        return None

    # if we fail to load the module, print some explanation
    # should not happen when run by endusers
    except BaseException:
        traceback.print_exc()
        raise



import asyncio
import os
import pkgutil
import signal
import sys
import time
from asyncio import create_subprocess_exec, create_subprocess_shell
from pathlib import PurePath
from threading import Thread
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from knot_resolver.utils.compat.asyncio import to_thread


def unblock_signals() -> None:
    if sys.version_info >= (3, 8):
        signal.pthread_sigmask(signal.SIG_UNBLOCK, signal.valid_signals())
    else:
        # the list of signals is not exhaustive, but it should cover all signals we might ever want to block
        signal.pthread_sigmask(
            signal.SIG_UNBLOCK,
            {
                signal.SIGHUP,
                signal.SIGINT,
                signal.SIGTERM,
                signal.SIGUSR1,
                signal.SIGUSR2,
            },
        )


async def call(
    cmd: Union[str, bytes, List[str], List[bytes]], shell: bool = False, discard_output: bool = False
) -> int:
    """Async alternative to subprocess.call()."""
    kwargs: Dict[str, Any] = {
        "preexec_fn": unblock_signals,
    }
    if discard_output:
        kwargs["stdout"] = asyncio.subprocess.DEVNULL
        kwargs["stderr"] = asyncio.subprocess.DEVNULL

    if shell:
        if isinstance(cmd, list):
            raise RuntimeError("can't use list of arguments with shell=True")
        proc = await create_subprocess_shell(cmd, **kwargs)
    else:
        if not isinstance(cmd, list):
            raise RuntimeError(
                "Please use list of arguments, not a single string. It will prevent ambiguity when parsing"
            )
        proc = await create_subprocess_exec(*cmd, **kwargs)

    return await proc.wait()


async def readfile(path: Union[str, PurePath]) -> str:
    """Asynchronously read whole file and return its content."""

    def readfile_sync(path: Union[str, PurePath]) -> str:
        with open(path, "r", encoding="utf8") as f:
            return f.read()

    return await to_thread(readfile_sync, path)


async def writefile(path: Union[str, PurePath], content: str) -> None:
    """Asynchronously set content of a file to a given string `content`."""

    def writefile_sync(path: Union[str, PurePath], content: str) -> int:
        with open(path, "w", encoding="utf8") as f:
            return f.write(content)

    await to_thread(writefile_sync, path, content)


async def wait_for_process_termination(pid: int, sleep_sec: float = 0) -> None:
    """
    Wait for the process termination.

    Will wait for any process (does not have to be a child process)
    given by its PID to terminate sleep_sec configures the granularity,
    with which we should return.
    """

    def wait_sync(pid: int, sleep_sec: float) -> None:
        while True:
            try:
                os.kill(pid, 0)
                if sleep_sec == 0:
                    os.sched_yield()
                else:
                    time.sleep(sleep_sec)
            except ProcessLookupError:
                break

    await to_thread(wait_sync, pid, sleep_sec)


async def read_resource(package: str, filename: str) -> Optional[bytes]:
    return await to_thread(pkgutil.get_data, package, filename)


T = TypeVar("T")


class BlockingEventDispatcher(Thread, Generic[T]):
    def __init__(self, name: str = "blocking_event_dispatcher") -> None:
        super().__init__(name=name, daemon=True)
        # warning: the asyncio queue is not thread safe
        self._removed_unit_names: "asyncio.Queue[T]" = asyncio.Queue()
        self._main_event_loop = asyncio.get_event_loop()

    def dispatch_event(self, event: T) -> None:
        """Dispatch events from the blocking thread."""

        async def add_to_queue() -> None:
            await self._removed_unit_names.put(event)

        self._main_event_loop.call_soon_threadsafe(add_to_queue)

    async def next_event(self) -> T:
        return await self._removed_unit_names.get()



# noqa: INP001
import argparse
import sys
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, comp_get_words, register_command
from knot_resolver.datamodel.cache_schema import CacheClearRPCSchema
from knot_resolver.utils.modeling.exceptions import AggregateDataValidationError, DataValidationError
from knot_resolver.utils.modeling.parsing import DataFormat, parse_json
from knot_resolver.utils.requests import request


class CacheOperations(Enum):
    CLEAR = 0


@register_command
class CacheCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)
        self.operation: Optional[CacheOperations] = namespace.operation if hasattr(namespace, "operation") else None
        self.output_format: DataFormat = (
            namespace.output_format if hasattr(namespace, "output_format") else DataFormat.YAML
        )

        # CLEAR operation
        self.clear_dict: Dict[str, Any] = {}
        if hasattr(namespace, "exact_name"):
            self.clear_dict["exact-name"] = namespace.exact_name
        if hasattr(namespace, "name"):
            self.clear_dict["name"] = namespace.name
        if hasattr(namespace, "rr_type"):
            self.clear_dict["rr-type"] = namespace.rr_type
        if hasattr(namespace, "chunk_size"):
            self.clear_dict["chunk-size"] = namespace.chunk_size

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        cache_parser = subparser.add_parser("cache", help="Performs operations on the cache of the running resolver.")

        config_subparsers = cache_parser.add_subparsers(help="operation type")

        # 'clear' operation
        clear_subparser = config_subparsers.add_parser(
            "clear", help="Purge cache records that match specified criteria."
        )
        clear_subparser.set_defaults(operation=CacheOperations.CLEAR, exact_name=False)
        clear_subparser.add_argument(
            "--exact-name",
            help="If set, only records with the same name are purged.",
            action="store_true",
            dest="exact_name",
        )
        clear_subparser.add_argument(
            "--rr-type",
            help="Optional, the resource record type to purge. It is supported only with the '--exact-name' flag set.",
            action="store",
            type=str,
        )
        clear_subparser.add_argument(
            "--chunk-size",
            help="Optional, the number of records to remove in one round; the default is 100."
            " The purpose is not to block the resolver for long."
            " The resolver repeats the cache clearing after one millisecond until all matching data is cleared.",
            action="store",
            type=int,
            default=100,
        )
        clear_subparser.add_argument(
            "name",
            type=str,
            nargs="?",
            help="Optional, subtree name to purge; if omitted,"
            " the entire cache is purged (and all other parameters are ignored).",
            default=None,
        )

        output_format = clear_subparser.add_mutually_exclusive_group()
        output_format_default = DataFormat.YAML
        output_format.add_argument(
            "--json",
            help="Set JSON as the output format.",
            const=DataFormat.JSON,
            action="store_const",
            dest="output_format",
            default=output_format_default,
        )
        output_format.add_argument(
            "--yaml",
            help="Set YAML as the output format. YAML is the default.",
            const=DataFormat.YAML,
            action="store_const",
            dest="output_format",
            default=output_format_default,
        )

        return cache_parser, CacheCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:
        if not self.operation:
            args.subparser.print_help()
            sys.exit()

        if self.operation == CacheOperations.CLEAR:
            try:
                validated = CacheClearRPCSchema(self.clear_dict)
            except (AggregateDataValidationError, DataValidationError) as e:
                print(e, file=sys.stderr)
                sys.exit(1)

            body: str = DataFormat.JSON.dict_dump(validated.get_unparsed_data())
            response = request(args.socket, "POST", "cache/clear", body)
            body_dict = parse_json(response.body)

        if response.status != 200:
            print(response, file=sys.stderr)
            sys.exit(1)
        print(self.output_format.dict_dump(body_dict, indent=4))



from typing import Any, Dict, List, Literal, Optional, Union

from knot_resolver.constants import WATCHDOG_LIB
from knot_resolver.datamodel.types import (
    DomainName,
    EscapedStr,
    IDPattern,
    IPAddress,
    ListOrItem,
    ReadableFile,
    TimeUnit,
)
from knot_resolver.utils.modeling import ConfigSchema


class RuleSchema(ConfigSchema):
    """
    Local data advanced rule configuration.

    ---
    name: Hostname(s).
    subtree: Type of subtree.
    address: Address(es) to pair with hostname(s).
    file: Path to file(s) with hostname and IP address(es) pairs in '/etc/hosts' like format.
    records: Direct addition of records in DNS zone file format.
    tags: Tags to link with other policy rules.
    ttl: Optional, TTL value used for these answers.
    nodata: Optional, use NODATA synthesis. NODATA will be synthesized for matching name, but mismatching type(e.g. AAAA query when only A exists).
    """  # noqa: E501

    name: Optional[ListOrItem[DomainName]] = None
    subtree: Optional[Literal["empty", "nxdomain", "redirect"]] = None
    address: Optional[ListOrItem[IPAddress]] = None
    file: Optional[ListOrItem[ReadableFile]] = None
    records: Optional[EscapedStr] = None
    tags: Optional[List[IDPattern]] = None
    ttl: Optional[TimeUnit] = None
    nodata: Optional[bool] = None
    # TODO: probably also implement the rule options from RPZSchema (.log + .dry_run)

    def _validate(self) -> None:
        options_sum = sum([bool(self.address), bool(self.subtree), bool(self.file), bool(self.records)])
        if options_sum == 2 and bool(self.address) and self.subtree in {"empty", "redirect"}:
            pass  # these combinations still make sense
        elif options_sum > 1:
            raise ValueError("only one of 'address', 'subtree' or 'file' can be configured")
        elif options_sum < 1:
            raise ValueError("one of 'address', 'subtree', 'file' or 'records' must be configured")

        options_sum2 = sum([bool(self.name), bool(self.file), bool(self.records)])
        if options_sum2 != 1:
            raise ValueError("one of 'name', 'file or 'records' must be configured")

        if bool(self.nodata) and bool(self.subtree) and not bool(self.address):
            raise ValueError("'nodata' defined but unused with 'subtree'")


class RPZSchema(ConfigSchema):
    class Raw(ConfigSchema):
        """
        Configuration or Response Policy Zone (RPZ).

        ---
        file: Path to the RPZ zone file.
        watchdog: Enables files watchdog for configured RPZ file. Requires the optional 'watchdog' dependency.
        tags: Tags to link with other policy rules.
        log: Enables logging information whenever this RPZ matches.
        """

        file: ReadableFile
        watchdog: Union[Literal["auto"], bool] = "auto"
        tags: Optional[List[IDPattern]] = None
        log: Optional[List[Literal["ip", "name"]]] = None
        # dry_run: bool = False

    _LAYER = Raw

    file: ReadableFile
    watchdog: bool
    tags: Optional[List[IDPattern]]
    log: Optional[List[Literal["ip", "name"]]]
    # dry_run: bool

    def _watchdog(self, obj: Raw) -> Any:
        if obj.watchdog == "auto":
            return WATCHDOG_LIB
        return obj.watchdog

    def _validate(self) -> None:
        if self.watchdog and not WATCHDOG_LIB:
            raise ValueError(
                "'watchdog' is enabled, but the required 'watchdog' dependency (optional) is not installed"
            )


class LocalDataSchema(ConfigSchema):
    """
    Local data for forward records (A/AAAA) and reverse records (PTR).

    ---
    ttl: Default TTL value used for added local data/records.
    nodata: Use NODATA synthesis. NODATA will be synthesized for matching name, but mismatching type(e.g. AAAA query when only A exists).
    addresses: Direct addition of hostname and IP addresses pairs.
    addresses_files: Direct addition of hostname and IP addresses pairs from files in '/etc/hosts' like format.
    records: Direct addition of records in DNS zone file format.
    rules: Local data rules.
    rpz: List of Response Policy Zones and its configuration.
    """  # noqa: E501

    ttl: Optional[TimeUnit] = None
    nodata: bool = True
    addresses: Optional[Dict[DomainName, ListOrItem[IPAddress]]] = None
    addresses_files: Optional[List[ReadableFile]] = None
    records: Optional[EscapedStr] = None
    rules: Optional[List[RuleSchema]] = None
    rpz: Optional[List[RPZSchema]] = None
    # root_fallback_addresses*: removed, rarely useful



import logging

from knot_resolver.controller.registered_workers import command_registered_workers
from knot_resolver.datamodel import KresConfig

logger = logging.getLogger(__name__)


async def files_reload(config: KresConfig, force: bool = False) -> None:
    cert_file = config.network.tls.cert_file
    key_file = config.network.tls.key_file

    if cert_file and key_file:
        if not cert_file.to_path().exists():
            logger.error(f"TLS cert files reload failed: cert-file {cert_file} file don't exist")
        elif not key_file.to_path().exists():
            logger.error(f"TLS cert files failed: cert-file {key_file} file don't exist")
        else:
            logger.info("TLS cert files reload triggered")
            cmd = f"net.tls('{cert_file}', '{key_file}')"
            await command_registered_workers(cmd)



from typing import Literal

# Policy actions
PolicyActionEnum = Literal[
    # Nonchain actions
    "pass",
    "deny",
    "drop",
    "refuse",
    "tc",
    "reroute",
    "answer",
    # Chain actions
    "mirror",
    "forward",
    "stub",
    "debug-always",
    "debug-cache-miss",
    "qtrace",
    "reqtrace",
]

# FLAGS from https://www.knot-resolver.cz/documentation/latest/lib.html?highlight=options#c.kr_qflags
PolicyFlagEnum = Literal[
    "no-minimize",
    "no-ipv4",
    "no-ipv6",
    "tcp",
    "resolved",
    "await-ipv4",
    "await-ipv6",
    "await-cut",
    "no-edns",
    "cached",
    "no-cache",
    "expiring",
    "allow_local",
    "dnssec-want",
    "dnssec-bogus",
    "dnssec-insecure",
    "dnssec-cd",
    "stub",
    "always-cut",
    "dnssec-wexpand",
    "permissive",
    "strict",
    "badcookie-again",
    "cname",
    "reorder-rr",
    "trace",
    "no-0x20",
    "dnssec-nods",
    "dnssec-optout",
    "nonauth",
    "forward",
    "dns64-mark",
    "cache-tried",
    "no-ns-found",
    "pkt-is-sane",
    "dns64-disable",
]

# DNS records from 'kres.type' table
DNSRecordTypeEnum = Literal[
    "A",
    "A6",
    "AAAA",
    "AFSDB",
    "ANY",
    "APL",
    "ATMA",
    "AVC",
    "AXFR",
    "CAA",
    "CDNSKEY",
    "CDS",
    "CERT",
    "CNAME",
    "CSYNC",
    "DHCID",
    "DLV",
    "DNAME",
    "DNSKEY",
    "DOA",
    "DS",
    "EID",
    "EUI48",
    "EUI64",
    "GID",
    "GPOS",
    "HINFO",
    "HIP",
    "HTTPS",
    "IPSECKEY",
    "ISDN",
    "IXFR",
    "KEY",
    "KX",
    "L32",
    "L64",
    "LOC",
    "LP",
    "MAILA",
    "MAILB",
    "MB",
    "MD",
    "MF",
    "MG",
    "MINFO",
    "MR",
    "MX",
    "NAPTR",
    "NID",
    "NIMLOC",
    "NINFO",
    "NS",
    "NSAP",
    "NSAP-PTR",
    "NSEC",
    "NSEC3",
    "NSEC3PARAM",
    "NULL",
    "NXT",
    "OPENPGPKEY",
    "OPT",
    "PTR",
    "PX",
    "RKEY",
    "RP",
    "RRSIG",
    "RT",
    "SIG",
    "SINK",
    "SMIMEA",
    "SOA",
    "SPF",
    "SRV",
    "SSHFP",
    "SVCB",
    "TA",
    "TALINK",
    "TKEY",
    "TLSA",
    "TSIG",
    "TXT",
    "UID",
    "UINFO",
    "UNSPEC",
    "URI",
    "WKS",
    "X25",
    "ZONEMD",
]



import enum
import logging
import os
import socket

logger = logging.getLogger(__name__)


class _Status(enum.Enum):
    NOT_INITIALIZED = 1
    FUNCTIONAL = 2
    FAILED = 3


_status = _Status.NOT_INITIALIZED
_socket = None


def systemd_notify(**values: str) -> None:
    global _status
    global _socket

    if _status is _Status.NOT_INITIALIZED:
        socket_addr = os.getenv("NOTIFY_SOCKET")
        os.unsetenv("NOTIFY_SOCKET")
        if socket_addr is None:
            _status = _Status.FAILED
            return
        if socket_addr.startswith("@"):
            socket_addr = socket_addr.replace("@", "\0", 1)

        try:
            _socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            _socket.connect(socket_addr)
            _status = _Status.FUNCTIONAL
        except Exception:
            _socket = None
            _status = _Status.FAILED
            logger.warning(f"Failed to connect to $NOTIFY_SOCKET at '{socket_addr}'", exc_info=True)
            return

    elif _status is _Status.FAILED:
        return

    if _status is _Status.FUNCTIONAL:
        assert _socket is not None
        payload = "\n".join((f"{key}={value}" for key, value in values.items()))
        try:
            _socket.send(payload.encode("utf8"))
        except Exception:
            logger.warning("Failed to send notification to systemd", exc_info=True)
            _status = _Status.FAILED
            _socket.close()
            _socket = None



# noqa: INP001
import argparse
from typing import List, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, comp_get_words, register_command


@register_command
class HelpCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)

    def run(self, args: CommandArgs) -> None:
        args.parser.print_help()

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        stop = subparser.add_parser("help", help="show this help message and exit")
        return stop, HelpCommand



# We disable pylint checks, because it can't find methods in newer Python versions.
#
# pylint: disable=no-member

import asyncio
import functools
import logging
import sys
from asyncio import AbstractEventLoop, coroutines, events, tasks
from typing import Any, Awaitable, Callable, Coroutine, Optional, TypeVar

logger = logging.getLogger(__name__)

T = TypeVar("T")


async def to_thread(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    # version 3.9 and higher, call directly
    if sys.version_info >= (3, 9):
        return await asyncio.to_thread(func, *args, **kwargs)  # type: ignore[attr-defined]

    # earlier versions, run with default executor
    loop = asyncio.get_event_loop()
    pfunc = functools.partial(func, *args, **kwargs)
    return await loop.run_in_executor(None, pfunc)


def async_in_a_thread(func: Callable[..., T]) -> Callable[..., Coroutine[None, None, T]]:
    async def wrapper(*args: Any, **kwargs: Any) -> T:
        return await to_thread(func, *args, **kwargs)

    return wrapper


def create_task(coro: Awaitable[T], name: Optional[str] = None) -> "asyncio.Task[T]":
    # version 3.8 and higher, call directly
    if sys.version_info >= (3, 8):
        # pylint: disable=unexpected-keyword-arg
        return asyncio.create_task(coro, name=name)  # type: ignore[attr-defined,arg-type,call-arg]

    # version 3.7 and higher, call directly without the name argument
    if sys.version_info >= (3, 8):
        return asyncio.create_task(coro)  # type: ignore[attr-defined,arg-type]

    # earlier versions, use older function
    return asyncio.ensure_future(coro)


def is_event_loop_running() -> bool:
    loop = events._get_running_loop()  # noqa: SLF001
    return loop is not None and loop.is_running()


def run(coro: Awaitable[T], debug: Optional[bool] = None) -> T:
    # Adapted version of this:
    # https://github.com/python/cpython/blob/3.9/Lib/asyncio/runners.py#L8

    # version 3.7 and higher, call directly
    # disabled due to incompatibilities
    if sys.version_info >= (3, 7):
        return asyncio.run(coro, debug=debug)  # type: ignore[attr-defined,arg-type]

    # earlier versions, use backported version of the function
    if events._get_running_loop() is not None:  # noqa: SLF001
        raise RuntimeError("asyncio.run() cannot be called from a running event loop")

    if not coroutines.iscoroutine(coro):
        raise ValueError(f"a coroutine was expected, got {repr(coro)}")

    loop = events.new_event_loop()
    try:
        events.set_event_loop(loop)
        if debug is not None:
            loop.set_debug(debug)
        return loop.run_until_complete(coro)
    finally:
        try:
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
            if hasattr(loop, "shutdown_default_executor"):
                loop.run_until_complete(loop.shutdown_default_executor())  # type: ignore[attr-defined]
        finally:
            events.set_event_loop(None)
            loop.close()


def _cancel_all_tasks(loop: AbstractEventLoop) -> None:
    # Backported from:
    # https://github.com/python/cpython/blob/3.9/Lib/asyncio/runners.py#L55-L74
    #
    to_cancel = tasks.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    if sys.version_info >= (3, 7):
        # since 3.7, the loop argument is implicitely the running loop
        # since 3.10, the loop argument is removed
        loop.run_until_complete(tasks.gather(*to_cancel, return_exceptions=True))
    else:
        loop.run_until_complete(tasks.gather(*to_cancel, loop=loop, return_exceptions=True))  # type: ignore[call-overload]

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def add_async_signal_handler(signal: int, callback: Callable[[], Coroutine[Any, Any, None]]) -> None:
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal, lambda: create_task(callback()))


def remove_signal_handler(signal: int) -> bool:
    loop = asyncio.get_event_loop()
    return loop.remove_signal_handler(signal)



from typing import Any, List, Literal, Optional, Union

from knot_resolver.constants import WATCHDOG_LIB
from knot_resolver.datamodel.types import (
    EscapedStr32B,
    Int0_512,
    Int0_65535,
    Int1_4096,
    InterfaceOptionalPort,
    IPAddress,
    IPAddressEM,
    IPNetwork,
    IPv4Address,
    IPv6Address,
    ListOrItem,
    PortNumber,
    ReadableFile,
    SizeUnit,
    WritableFilePath,
)
from knot_resolver.utils.modeling import ConfigSchema

KindEnum = Literal["dns", "xdp", "dot", "doh-legacy", "doh2", "doq"]


class EdnsBufferSizeSchema(ConfigSchema):
    """
    EDNS payload size advertised in DNS packets.

    ---
    upstream: Maximum EDNS upstream (towards other DNS servers) payload size.
    downstream: Maximum EDNS downstream (towards clients) payload size for communication.
    """

    upstream: SizeUnit = SizeUnit("1232B")
    downstream: SizeUnit = SizeUnit("1232B")


class AddressRenumberingSchema(ConfigSchema):
    """
    Renumbers addresses in answers to different address space.

    ---
    source: Source subnet.
    destination: Destination address prefix.
    """

    source: IPNetwork
    destination: Union[IPAddressEM, IPAddress]


class QUICSchema(ConfigSchema):
    """
    Optional DoQ configuration.

    ---
    max_conns: Maximum number of active connections a single worker is allowed to accept.
    max_streams: Maximum number of concurrent streams a connection is allowed to open.
    require_retry: Require address validation for unknown source addresses.
                    This adds a 1-RTT delay to connection establishment.
    """

    max_conns: Int1_4096 = Int1_4096(1024)
    max_streams: Int1_4096 = Int1_4096(1024)
    require_retry: bool = False


class TLSSchema(ConfigSchema):
    class Raw(ConfigSchema):
        """
        TLS configuration, also affects DNS over TLS and DNS over HTTPS.

        ---
        watchdog: Enables watchdog of changes in TLS certificate files. Requires the optional 'watchdog' dependency.
        cert_file: Path to certificate file.
        key_file: Path to certificate key file.
        sticket_secret: Secret for TLS session resumption via tickets. (RFC 5077).
        sticket_secret_file: Path to file with secret for TLS session resumption via tickets. (RFC 5077).
        padding: EDNS(0) padding of queries and answers sent over an encrypted channel.
        """

        watchdog: Union[Literal["auto"], bool] = "auto"
        cert_file: Optional[ReadableFile] = None
        key_file: Optional[ReadableFile] = None
        sticket_secret: Optional[EscapedStr32B] = None
        sticket_secret_file: Optional[ReadableFile] = None
        padding: Union[bool, Int0_512] = True

    _LAYER = Raw

    watchdog: bool
    cert_file: Optional[ReadableFile] = None
    key_file: Optional[ReadableFile] = None
    sticket_secret: Optional[EscapedStr32B] = None
    sticket_secret_file: Optional[ReadableFile] = None
    padding: Union[bool, Int0_512] = True

    def _watchdog(self, obj: Raw) -> Any:
        if obj.watchdog == "auto":
            return WATCHDOG_LIB
        return obj.watchdog

    def _validate(self) -> None:
        if self.sticket_secret and self.sticket_secret_file:
            raise ValueError("'sticket_secret' and 'sticket_secret_file' are both defined, only one can be used")
        if bool(self.cert_file) != bool(self.key_file):
            raise ValueError("'cert-file' and 'key-file' must be configured together")
        if self.cert_file and self.key_file and self.watchdog and not WATCHDOG_LIB:
            raise ValueError(
                "'files-watchdog' is enabled, but the required 'watchdog' dependency (optional) is not installed"
            )


class ListenSchema(ConfigSchema):
    class Raw(ConfigSchema):
        """
        Configuration of listening interface.

        ---
        unix_socket: Path to unix domain socket to listen to.
        interface: IP address or interface name with optional port number to listen to.
        port: Port number to listen to.
        kind: Specifies DNS query transport protocol.
        freebind: Used for binding to non-local address.
        """

        interface: Optional[ListOrItem[InterfaceOptionalPort]] = None
        unix_socket: Optional[ListOrItem[WritableFilePath]] = None
        port: Optional[PortNumber] = None
        kind: KindEnum = "dns"
        freebind: bool = False

    _LAYER = Raw

    interface: Optional[ListOrItem[InterfaceOptionalPort]]
    unix_socket: Optional[ListOrItem[WritableFilePath]]
    port: Optional[PortNumber]
    kind: KindEnum
    freebind: bool

    def _interface(self, origin: Raw) -> Optional[ListOrItem[InterfaceOptionalPort]]:
        if origin.interface:
            port_set: Optional[bool] = None
            for intrfc in origin.interface:  # type: ignore[attr-defined]
                if origin.port and intrfc.port:
                    raise ValueError("The port number is defined in two places ('port' option and '@<port>' syntax).")
                if port_set is not None and (bool(intrfc.port) != port_set):
                    raise ValueError(
                        "The '@<port>' syntax must be used either for all or none of the interface in the list."
                    )
                port_set = bool(intrfc.port)
        return origin.interface

    def _port(self, origin: Raw) -> Optional[PortNumber]:
        if origin.port:
            return origin.port
        # default port number based on kind
        if origin.interface:
            if origin.kind in ["dot", "doq"]:
                return PortNumber(853)
            if origin.kind in ["doh-legacy", "doh2"]:
                return PortNumber(443)
            return PortNumber(53)
        return None

    def _validate(self) -> None:
        if bool(self.unix_socket) == bool(self.interface):
            raise ValueError("One of 'interface' or 'unix-socket' must be configured.")
        if self.port and self.unix_socket:
            raise ValueError(
                "'unix-socket' and 'port' are not compatible options."
                " Port configuration can only be used with 'interface' option."
            )


class ProxyProtocolSchema(ConfigSchema):
    """
    PROXYv2 protocol configuration.

    ---
    enable: Enable/disable PROXYv2 protocol.
    allow: Allow usage of the PROXYv2 protocol headers by clients on the specified addresses.
    """

    enable: bool = False
    allow: Optional[List[Union[IPAddress, IPNetwork]]] = None


class NetworkSchema(ConfigSchema):
    """
    Network connections and protocols configuration.

    ---
    do_ipv4: Enable/disable using IPv4 for contacting upstream nameservers.
    do_ipv6: Enable/disable using IPv6 for contacting upstream nameservers.
    out_interface_v4: IPv4 address used to perform queries. Not set by default, which lets the OS choose any address.
    out_interface_v6: IPv6 address used to perform queries. Not set by default, which lets the OS choose any address.
    tcp_pipeline: TCP pipeline limit. The number of outstanding queries that a single client connection can make in parallel.
    edns_tcp_keepalive: Allows clients to discover the connection timeout. (RFC 7828)
    edns_buffer_size: Maximum EDNS payload size advertised in DNS packets. Different values can be configured for communication downstream (towards clients) and upstream (towards other DNS servers).
    address_renumbering: Renumbers addresses in answers to different address space.
    tls: TLS configuration, also affects DNS over TLS, DNS over HTTPS and DNS over QUIC.
    quic: DNS over QUIC configuration.
    proxy_protocol: PROXYv2 protocol configuration.
    listen: List of interfaces to listen to and its configuration.
    """  # noqa: E501

    do_ipv4: bool = True
    do_ipv6: bool = True
    out_interface_v4: Optional[IPv4Address] = None
    out_interface_v6: Optional[IPv6Address] = None
    tcp_pipeline: Int0_65535 = Int0_65535(100)
    edns_tcp_keepalive: bool = True
    edns_buffer_size: EdnsBufferSizeSchema = EdnsBufferSizeSchema()
    address_renumbering: Optional[List[AddressRenumberingSchema]] = None
    tls: TLSSchema = TLSSchema()
    proxy_protocol: ProxyProtocolSchema = ProxyProtocolSchema()
    quic: QUICSchema = QUICSchema()
    listen: List[ListenSchema] = [
        ListenSchema({"interface": "127.0.0.1"}),
        ListenSchema({"interface": "::1", "freebind": True}),
    ]
